import os
import rasterio

import numpy as np


class wp_demographics(object):
    def __init__(self, wp_file):
        self.wp_file = wp_file
        self.age = os.path.basename(wp_file).split("_")[2]
        self.gender = os.path.basename(wp_file).split("_")[1]
        self.country = os.path.basename(wp_file).split("_")[0]
        self.raster = rasterio.open(wp_file)

    def __str__(self):
        return f"{self.country} - {self.gender} - {self.age}"


class wp_vulnerability(object):
    def __init__(self, wp_files, vulnerability_definition):
        """Process the wp demographic information into a single map of vulnerability

        INPUT
        wp_files [list of wp_demographics]
        vulnerability [dictionary of definitions]
        """
        self.wp_files = wp_files
        self.v_def = vulnerability_definition
        # classify wp_files into categories
        vul_files = {}
        for key, value in vulnerability_definition.items():
            min_val = int(key.split("-")[0])
            max_val = int(key.split("-")[1])
            for f in wp_files:
                # print(f'{f.age}: {min_val}')
                if (int(f.age) > min_val) and (int(f.age) <= max_val):
                    try:
                        vul_files[key].append(f)
                    except:
                        vul_files[key] = [f]
        self.vul_files = vul_files

    def calculate_total_pop(self):
        vul_results = {}
        for key, values in self.vul_files.items():
            for wp in values:
                curR = wp.raster.read()
                try:
                    final = final + curR
                except:
                    final = curR
        self.total_pop = final

    def calculate_vulnerability(self):
        vul_results = {}
        for key, values in self.vul_files.items():
            vul_value = self.v_def[key]
            for wp in values:
                curR = wp.raster.read()
                curR = curR * vul_value
                try:
                    vul_results[key] = vul_results[key] + curR
                except:
                    vul_results[key] = curR
        self.vul_res = vul_results

    def combine_results(self, out_vulnerability="", out_pop=""):
        for key, values in self.vul_res.items():
            for v in values:
                try:
                    final = final + v
                except:
                    final = v
        final[final < 0] = np.nan
        if out_vulnerability != "":
            with rasterio.open(
                out_vulnerability, "w", **self.wp_files[0].raster.meta
            ) as outR:
                outR.write_band(1, final)
        else:
            return final

        if out_pop != "":
            with rasterio.open(out_pop, "w", **self.wp_files[0].raster.meta) as outR:
                outR.write(self.total_pop)



import json
import os
import time
import logging

import pandas as pd
import urllib.request as url

### README
### This code has been adpated from the earlier Market Access tools written by Charles Fox.
### Some simple adjustments were made to work with the InfraSAP inputs, but I have not had enough time to re-write this as it's quite complicated.


def CreateODMatrix(
    input_df,
    input_df2,
    lat_name="Lat",
    lon_name="Lon",
    UID="ID",
    Pop="Pop",
    call_type="OSRM",
    rescue=0,
    rescue_num=0,
    MB_Toke="",
    sleepTime=5,
    osrmHeader="",
):
    """
    TODO: make the function flexible for MapBox endpoint - add back MB token / different formatting
    """
    #     ffpath = os.path.dirname(infile)
    start = time.time()
    print("\nChosen server: %s\n\nStart time: %s" % (call_type, time.ctime(start)))
    #     print('Origins: %s' % infile)
    #     print('Destinations: %s\n' % infile_2)

    # Save settings
    save_rate = 5

    def save(returns, j, i, numcalls, rescue_num):
        elapsed_mins = (time.time() - start) / 60
        elapsed_secs = (time.time() - start) % 60
        total = (numcalls / float(i)) * (time.time() - start) / 60.0
        remaining = total - elapsed_mins
        print("\n______________________________________\n")
        print(
            "\nSave point %s. Running for: %d minutes %d seconds"
            % (j, elapsed_mins, elapsed_secs)
        )
        print(
            "\ncalls completed: %d of %d. Est. run time: %d minutes. Time remaining: %d"
            % (i - 1, numcalls, total, remaining)
        )
        print("\npercentage complete: %d percent" % ((i - 1) / float(numcalls) * 100))
        print("\n______________________________________\n")
        try:
            df = pd.concat(returns)
        except:
            df = returns
        curOutput = os.path.join(ffpath, "temp_file_%d.csv" % rescue_num)
        df.to_csv(curOutput)

    # Function for calling OSRM server.
    def Call(O_list, D_list, i, O_IDs, D_IDs, header):
        # Convert origins to HTTP request string
        Os = ";".join(str(coord).replace("'", "").replace(";", "") for coord in O_list)
        # Destinations to HTTP request string
        Ds = ";".join(str(coord).replace("'", "").replace(";", "") for coord in D_list)
        # Join them together
        data = Os + ";" + Ds

        # Define which coords in data string are origins, and which are destinations
        sources = ["%d" % x for x in range(0, len(O_list))]
        sources = ";".join(str(x).replace("'", "") for x in sources)
        lenth = len(O_list) + len(D_list)
        destinations = ["%d" % x for x in range(len(O_list), lenth)]
        destinations = ";".join(str(x).replace("'", "") for x in destinations)

        # Build request string
        request = (
            header
            + data
            + "?sources="
            + sources
            + "&destinations="
            + destinations
            + "?&access_token="
            + MB_Toke
        )
        # Pass request to interweb

        try:
            r = url.urlopen(request)
        except:
            print(request)
            time.sleep(5)
            r = url.urlopen(request)

        # Error handle
        try:
            # Convert Bytes response to readable Json
            MB_TelTest_json = json.loads(r.read().decode("utf-8"))
            data_block = MB_TelTest_json["durations"]
        except:
            data_block = "null"

        # Build df from JSON
        # sources_label = [str(i['location']) for i in MB_TelTest_json['sources']]
        # dest_label = [str(i['location']) for i in MB_TelTest_json['destinations']]
        sources_label = O_IDs
        dest_label = D_IDs
        chunk = pd.DataFrame(data=data_block, columns=dest_label, index=sources_label)
        # Convert to minutes, stack 2D array to 1D array
        chunk = chunk.stack(level=-1)
        chunk.columns = ["O", "D", "DIST"]
        return chunk

    # Generate appropriately split source and destination lists
    def split_and_bundle(in_list, break_size):
        new_list = []
        for i in range(0, (int(max(len(in_list) / break_size, 1)))):
            upper = (i + 1) * break_size
            lower = upper - break_size
            objs = in_list[lower:upper]
            new_list.append(objs)
        if len(in_list) > break_size:
            rem = len(in_list) % break_size
            if rem > 0:
                final = upper + rem
                new_list.append(in_list[upper:final])
        return new_list

    # File Import for sources file
    #     input_df = pd.read_csv(infile)
    input_df["source_list"] = (
        input_df[lon_name].map(str).str.cat(input_df[lat_name].map(str), sep=",")
    )
    input_df["source_list"] = input_df["source_list"] + ";"
    source_list = input_df["source_list"].values.tolist()
    source_UIDs = input_df[UID].values.tolist()
    # input_df['source_point'] = input_df.apply(lambda x: Point(x[lon_name],x[lat_name]), axis = 1)
    # source_points = input_df['source_point'].tolist()

    # Look to import separate file for destinations; if not, set destinations = sources
    #     input_df2 = pd.read_csv(infile_2)
    input_df2["dest_list"] = (
        input_df2[lon_name].map(str).str.cat(input_df2[lat_name].map(str), sep=",")
    )
    input_df2["dest_list"] = input_df2["dest_list"] + ";"
    dest_list = input_df2["dest_list"].values.tolist()
    dest_UIDs = input_df2[UID].values.tolist()

    if call_type == "MBT":
        sources_list = split_and_bundle(source_list, 5)
        dests_list = split_and_bundle(dest_list, 5)
        sources_UIDs = split_and_bundle(source_UIDs, 5)
        dests_UIDs = split_and_bundle(dest_UIDs, 5)
    elif call_type == "MB" or call_type == "OSRM":
        sources_list = split_and_bundle(source_list, 12)
        dests_list = split_and_bundle(dest_list, 13)
        sources_UIDs = split_and_bundle(source_UIDs, 12)
        dests_UIDs = split_and_bundle(dest_UIDs, 13)
    else:
        pass

    # Run function call across the O-D matrix; output is 'df'
    returns = []
    numcalls = len(sources_list) * len(dests_list)
    s, d = sources_list, dests_list
    i, j = 1 + (rescue * len(sources_list)), 1 + rescue

    ### Making Calls
    if call_type == "Euclid":
        df = EuclidCall(source_list, dest_list, source_points, dest_points)
    else:
        if rescue > 0:
            s = s[rescue:]  # possibly rescue -1
            sources_UIDs = sources_UIDs[rescue:]
        print("source list: %s" % len(source_list))
        print("sources list: %s" % len(sources_list))
        print("dest list: %s" % len(dest_list))
        print("dests list: %s" % len(dests_list))
        numcalls_rem = len(s) * len(d)
        print("\nEstimated remaining calls to chosen server: %d\n" % numcalls_rem)
        print("save points will occur every %d calls\n" % (len(dests_list)))
        if sleepTime > 0:
            time.sleep(sleepTime)
        for O_list in s:
            O_IDs = sources_UIDs[s.index(O_list)]
            for D_list in d:
                if sleepTime > 0:
                    time.sleep(sleepTime)
                D_IDs = dests_UIDs[d.index(D_list)]
                if call_type == "MB":
                    header = (
                        "https://api.mapbox.com/directions-matrix/v1/mapbox/driving/"
                    )
                elif call_type == "MBT":
                    header = "https://api.mapbox.com/directions-matrix/v1/mapbox/driving-traffic/"
                elif call_type == "OSRM":
                    header = "http://router.project-osrm.org/table/v1/driving/"
                    if osrmHeader != "":
                        header = osrmHeader
                try:
                    # prevent server annoyance
                    print("Call to OSRM server number: %d of %s" % (i, numcalls_rem))
                    returns.append(Call(O_list, D_list, i, O_IDs, D_IDs, header))
                    i += 1
                    j += 1
                except:
                    logging.warning("Error Processing OSRM for i:%s and j:%s" % (i, j))
                    save(returns, j, i, numcalls, rescue_num)
        try:
            df = pd.concat(returns)
        except:
            df = returns

    # re-attach the population of origins and destinations, prep dataframe
    all_matrices = []
    if rescue_num > 0:
        for r in range(0, rescue_num):
            rescued_matrix = pd.read_csv(
                os.path.join(ffpath, "temp_file_%d.csv" % (r)), header=None
            )
            rescued_matrix.columns = ["O_UID", "D_UID", "DIST"]
            all_matrices.append(rescued_matrix)
    df = df.reset_index()
    df.columns = ["O_UID", "D_UID", "DIST"]
    all_matrices.append(df)
    new = pd.concat(all_matrices)
    new = new.set_index("O_UID")
    new["DIST"] = new["DIST"].apply(pd.to_numeric)
    popdf = input_df[[UID, Pop]].set_index(UID)
    new["O_POP"] = popdf[Pop]
    new = new.reset_index()
    new = new.set_index("D_UID")
    if dest_list == source_list:
        new["D_POP"] = popdf[Pop]
        new = new.reset_index()
    else:
        popdf_dest = input_df2[[UID, Pop]].set_index(UID)
        new["D_POP"] = popdf_dest[Pop]
        new = new.reset_index()
    new["O_UID"] = new["O_UID"].astype(str)
    new["D_UID"] = new["D_UID"].astype(str)
    new["combo"] = new["O_UID"] + "_X_" + new["D_UID"]
    new = new.drop_duplicates("combo")
    new = new.drop(["combo"], axis=1)

    return new



import geopandas as gpd
import pandas as pd

from shapely.geometry import Point


def extract_power_plants(power_plants_database, country_bounds):
    pp_d = pd.read_csv(power_plants_database)
    pp_geom = [Point(x) for x in zip(pp_d["longitude"], pp_d["latitude"])]
    pp_d = gpd.GeoDataFrame(pp_d, geometry=pp_geom, crs={"init": "epsg:4326"})

    selected_pp = pp_d[pp_d.intersects(country_bounds.unary_union)]
    return selected_pp


def extract_transmission_lines(grid_line_file, out_bounds):
    lines = gpd.read_file(grid_line_file)
    sel_lines = lines[lines.intersects(out_bounds.unary_union)]
    return sel_lines



import os
import subprocess
import copy
import multiprocessing
import rasterio
import geohash

import geopandas as gpd
import pandas as pd

from collections import Counter
from multiprocessing import Pool

try:
    from . import vulnerability_mapping as vulmap
    from . import misc
    from . import rasterMisc as rMisc
    from . import UrbanRaster as urban
except:
    import vulnerability_mapping as vulmap
    import misc
    import rasterMisc as rMisc
    import UrbanRaster as urban
"""
import vulnerability_mapping as vulmap
import misc as misc
import osmMisc as osm
import UrbanRaster as urban
"""
# https://mrc-ide.github.io/global-lmic-reports/parameters.html
# https://www.cdc.gov/mmwr/volumes/69/wr/mm6912e2.htm
vul_def = {
    "0-5": 0.001,
    "6-10": 0.001,
    "11-15": 0.001,
    "16-20": 0.002,
    "21-25": 0.005,
    "26-30": 0.010,
    "31-35": 0.016,
    "36-40": 0.023,
    "41-45": 0.029,
    "46-50": 0.039,
    "51-55": 0.058,
    "56-60": 0.072,
    "61-65": 0.102,
    "66-70": 0.117,
    "71-75": 0.146,
    "76-80": 0.177,
    "81-100": 0.180,
}

hnp_categories = {
    "R10": {
        "Name": "WorldPop",
        "raster_file": "WP_2020_1km.tif",
        "vars": ["SUM"],
        "description": "WorldPop total population in 2020",
    },
    "P1": {
        "Name": "Urban_Population",
        "raster_file": "WP_2020_1km_urban_pop.tif",
        "vars": ["SUM"],
        "description": "WorldPop total urban population in 2020",
    },
    "P2": {
        "Name": "Demographics",
        "raster_file": "WP2020_vulnerability_map.tif",
        "vars": ["SUM"],
        "description": "Total potential hospitalization load based on WorldPop demographics and published CoVID hospitalization rates",
    },
    "LC": {
        "Name": "Landcover",
        "raster_file": "LC.tif",
        "vars": ["C"],
        "unqVals": [
            11,
            14,
            20,
            30,
            40,
            50,
            60,
            70,
            90,
            100,
            110,
            120,
            130,
            140,
            150,
            160,
            170,
            180,
            190,
            200,
            210,
            220,
            230,
        ],
        "description": "Landcover dataset from Globcover",
    },
}


def create_fishnet(extents_file, out_folder, prefix, verbose=True):
    """Create a 1 km fishnet inside each feature in the input extents_file

    INPUT
        extents_file [string] - path to urban extents
        out_folder [string path] - where output shapefiles should be written
        prefix [string] - will be appended to each fidhnet shapefile
    """
    urban_extents = gpd.read_file(extents_file)
    # sel_cities = urban_extents.sort_values(['Pop'], ascending=False).iloc[0:5]
    sel_cities = urban_extents.sort_values(["Pop"], ascending=False)
    try:
        sel_cities = misc.project_UTM(sel_cities)
    except:
        sel_cities = sel_cities.to_crs({"init": "epsg:3857"})

    for idx, row in sel_cities.iterrows():
        out_fishnet = os.path.join(out_folder, "%s_%s.shp" % (prefix, row["ID"]))
        if not os.path.exists(out_fishnet):
            b = row["geometry"].bounds
            crs_num = sel_cities.crs["init"].split(":")[-1]
            crs_num = int(crs_num)
            misc.createFishnet(
                out_fishnet, b[0], b[2], b[1], b[3], 1000, 1000, crsNum=crs_num
            )
            fishnet = gpd.read_file(out_fishnet)
            fishnet = fishnet[fishnet.intersects(row["geometry"])]
            fishnet = fishnet.to_crs({"init": "epsg:4326"})
            fishnet["geohash"] = fishnet["geometry"].apply(
                lambda x: geohash.encode(x.centroid.y, x.centroid.x)
            )
            fishnet.to_file(out_fishnet)
            if verbose:
                misc.tPrint("%s: %s" % (prefix, row["ID"]))


def create_urban_data(
    iso3,
    country_folder,
    country_bounds,
    inR,
    calc_urban=True,
    calc_hd_urban=True,
    verbose=False,
    urb_dens=300,
    urb_pop=5000,
    hd_urb_dens=1500,
    hd_urb_pop=50000,
):
    """Extract urban areas from gridded population data following the EC density methodology

    INPUTS:
        iso3 [string] - iso3 code for country of interest
        country_folder [string] - path to output folder
        country_bounds [geopandas dataframe] - boundary within which to extract
        inR [rasterio object] - population datasets from which country specific pop layer is extracted
        [optional] calc_urban [boolean] - whether to calculate urban (lower density) extents
        [optional] calc_hd_urban [boolean] - whether to calculate hd urban (higher density) extents
        [optional] urb_dens/hd_urb_dens [int] - population density threshold for calculating urban areas IN THE PER PIXEL UNITS OF inR
        [optional] urb_pop/hd_urb_pop [int] - total population threshold for calculating urban areas

    RETURNS:
        NA - the function calculates a number of files in the folder country_folder:
            -- WP_2020_1km.tif: population dataaset clipped from inR
            -- urban_areas.shp: vectorization of urban areas calculation
            -- urban_areas_hd.shp: vectorization of high density urban areas calculation
            -- urban_fishnets/URBAN_XX.shp: a 1 km fishnet is created over the five highest population areas in urban_areas.shp
            -- hd_urban_fishnets/HD_URBAN_XX.shp: a 1 km fishnet is created over the five highest population areas in hd_urban_areas.shp
    """

    country_pop = os.path.join(country_folder, "WP_2020_1km.tif")
    urban_pop = os.path.join(country_folder, "WP_2020_1km_urban_pop.tif")
    hd_urban_pop = os.path.join(country_folder, "WP_2020_1km_hd_urban_pop.tif")
    urb_bounds = os.path.join(country_folder, "urban_areas.shp")
    hd_urb_bounds = os.path.join(country_folder, "urban_areas_hd.shp")
    urban_fishnets = os.path.join(country_folder, "urban_fishnets")
    hd_urban_fishnets = os.path.join(country_folder, "hd_urban_fishnets")

    # Clip pop raster
    if not os.path.exists(country_pop):
        rMisc.clipRaster(inR, country_bounds, country_pop)

    if not os.path.exists(urban_pop) and calc_urban:
        urbanR = urban.urbanGriddedPop(country_pop)
        urban_vec = urbanR.calculateUrban(
            densVal=urb_dens,
            totalPopThresh=urb_pop,
            smooth=True,
            queen=False,
            raster_pop=urban_pop,
        )
        urban_vec.to_file(urb_bounds)

    if not os.path.exists(hd_urban_pop) and calc_hd_urban:
        urbanR = urban.urbanGriddedPop(country_pop)
        urban_vec = urbanR.calculateUrban(
            densVal=hd_urb_dens,
            totalPopThresh=hd_urb_pop,
            smooth=True,
            queen=True,
            raster_pop=urban_pop,
        )
        urban_vec.to_file(hd_urb_bounds)
    # Generate Fishnets
    if not os.path.exists(urban_fishnets):
        os.makedirs(urban_fishnets)

    if not os.path.exists(hd_urban_fishnets):
        os.makedirs(hd_urban_fishnets)

    if calc_urban:
        create_fishnet(urb_bounds, urban_fishnets, "URBAN")
    if calc_hd_urban:
        create_fishnet(hd_urb_bounds, hd_urban_fishnets, "HD_URBAN")


def calculate_vulnerability(
    iso3, country_folder, country_bounds, pop_folder, pop_files
):
    """The hospitalization rates listed in the vul_def dictionary (at the top of this script) are
    combined to create a vulnerability layer

    INPUTS
        iso3 [string] - iso3 code for country of interest
        country_folder [string] - path to output folder
        country_bounds [geopandas dataframe] - boundary within which to extract
        pop_files [list of file paths] - list of global demographic rasters which are clipped out with country_bounds
    RETURNS
        NA - creates a file called WP2020_vulnerability_map.tif in country_folder
    """

    # clip out the temporary vulnerability metrics
    out_vulnerability = os.path.join(country_folder, "WP2020_vulnerability_map.tif")
    if not os.path.exists(out_vulnerability):
        wp_files = []
        if not os.path.exists(country_folder):
            os.makedirs(country_folder)
        for pFile in pop_files:
            curR = rasterio.open((os.path.join(pop_folder, pFile)))
            out_file = os.path.join(country_folder, pFile)
            if not os.path.exists(out_file):
                rMisc.clipRaster(curR, country_bounds, out_file)
            wp_files.append(out_file)
        # Calculate vulnerability
        wp_file_objects = [
            vulmap.wp_demographics(os.path.join(country_folder, x)) for x in wp_files
        ]
        vul = vulmap.wp_vulnerability(wp_file_objects, vul_def)
        vul.calculate_vulnerability()
        vul.combine_results(out_vulnerability, "")
        for f in wp_files:
            os.remove(f)


def extract_osm(
    sel_country,
    out_folder,
    global_osm_pbf="/home/public/Data/GLOBAL/OSM/GLOBAL/planet-latest.osm.pbf",
):
    """Extract a national osm pbf from the global file
    INPUTS
        sel_country [shapely object] - extent to clip
        out_folder [string folder path] - output folder for creating the osm.osm.pbf
    """

    osm_extractor = osm.osmExtraction(
        osmosisCmd="/home/wb411133/Code/Osmosis/bin/osmosis",
        tempFile="/home/wb411133/temp/osmosis.sh",
    )
    out_pbf = os.path.join(out_folder, "osm.osm.pbf")
    out_roads = os.path.join(out_folder, "OSM_roads.shp")
    if not os.path.exists(out_pbf):
        cmds = osm_extractor.extractBoundingBox(
            global_osm_pbf, sel_country, out_pbf, execute=False
        )
        subprocess.check_call(cmds.split(" "))
    if not os.path.exists(out_roads):
        roads = osm.convertOSMPBF_DataFrame(out_pbf, "lines")
        roads_sel = roads.loc[:, ["geometry", "highway", "Length", "osm_id"]]
        roads_sel = roads_sel.loc[[x is not None for x in roads["highway"]]]
        roads_sel.to_file(out_roads)


def run_zonal(admin_shapes, rasters, out_suffix="", iso3=""):
    """Calculate zonal results for submitted admin and raster

    INPUTS
        admin_shapes [geopandas] - features within which to calculate statistics
        rasters [dictionary] - data dictionary containing the raster and the required information
            { 'HNP_Var1':{
                    'raster_file': 'path_to_raster',
                    'vars':['SUM','MEAN'],
                    'description':'Lorem Ipsum'
                }
            }
        out_suffix [string] - text to append to output zonal file
    """
    for shp in admin_shapes:
        inD = gpd.read_file(shp)
        out_zonal = shp.replace(".shp", "_zonal%s.csv" % out_suffix)
        misc.tPrint(f"Processed: {iso3} {os.path.basename(shp)}")
        write_out = False
        if not os.path.exists(out_zonal):
            for var_name, definition in rasters.items():
                if os.path.exists(definition["raster_file"]):
                    write_out = True
                    if definition["vars"][0] == "C":
                        uVals = definition["unqVals"]
                        res = rMisc.zonalStats(
                            inD,
                            definition["raster_file"],
                            rastType="C",
                            unqVals=uVals,
                            reProj=True,
                        )
                        res = pd.DataFrame(res, columns=["LC_%s" % x for x in uVals])
                        for column in res.columns:
                            inD[column] = res[column]
                    else:
                        # Zonal stats
                        res = rMisc.zonalStats(
                            inD, definition["raster_file"], minVal=0, reProj=True
                        )
                        res = pd.DataFrame(res, columns=["SUM", "MIN", "MAX", "MEAN"])
                        res.columns = [f"{var_name}_{x}" for x in res.columns]
                        for var in definition["vars"]:
                            inD[f"{var_name}_{var}"] = res[f"{var_name}_{var}"]
            if write_out:
                inD.drop(["geometry"], axis=1, inplace=True)
                pd.DataFrame(inD).to_csv(out_zonal)


def check_zonal(country_folder, remove_bad=False):
    """Check the results of the zonal statistics"""

    stat_files = []
    for root, dirs, files in os.walk(country_folder):
        if "FINAL" not in root:
            for f in files:
                if f[-4:] == ".csv":
                    stat_files.append(os.path.join(root, f))

    res = {}
    for stat_file in stat_files:
        if "DHS" in stat_file:
            good_col = "age_final_0_4_househ_SUM"
            bad_col = "R10_SUM"
        else:
            good_col = "R10_SUM"
            bad_col = "age_final_0_4_househ_SUM"
        xx = pd.read_csv(stat_file)
        cols = list(xx.columns)
        val = 0
        if good_col in cols:
            val += 1
        if bad_col in cols:
            val += 10
        res[stat_file] = val

    if remove_bad:
        for f, score in res.items():
            if score in [0, 10]:
                os.remove(f)

    return res


def combine_dhs_pop(popRaster, dhs_raster, out_file, factor=100):
    """
    INPUT
        popRaster [rasterio]
        dhs_raster [rasterio]
        out_file [string]
        [optional] factor [int] - number to divide dhs_raster (converts percentage to fraction)
    """
    inP = popRaster.read()
    dhs = dhs_raster.read()
    if factor != 1:
        dhs = dhs / factor
    dhs_pop = inP * (dhs)

    with rasterio.open(out_file, "w", **popRaster.meta) as outR:
        outR.write(dhs_pop)


def summarize_DHS(template, dhs_files, country_folder, iso3):
    """combine DHS data with WorldPop population and run zonal stats

    INPUT
        template [string] - template raster upon which to base rasterization of DHS
        dhs_files [dictionary] - defines DHS files to process {filename:geopandas}
        country_folder [string to path] - folder to create output
    """
    # Process DHS data
    inP = rasterio.open(template)
    # get a list of unique columns in the DHS data
    total_columns = 0
    try:
        del all_columns
    except:
        pass
    # get a list of all unique columns
    for key, inD in dhs_files.items():
        cur_columns = list(inD.columns.values)
        try:
            all_columns = all_columns + cur_columns
        except:
            all_columns = cur_columns

    col_count = Counter(all_columns)
    unq_columns = [key for key, value in col_count.items() if value == 1]

    dhs_rasters = {}
    for key, inD in dhs_files.items():
        sel_dhs = inD.loc[inD["ISO3"] == iso3]
        if sel_dhs.shape[0] > 0:
            for field in inD.columns:
                if field in unq_columns:
                    out_file = os.path.join(country_folder, f"{key}_{field}.tif")
                    out_file_pop = os.path.join(
                        country_folder, f"{key}_{field}_pop.tif"
                    )
                    try:
                        # rasterize the desired field in the inputDHS data
                        if not os.path.exists(out_file) and not os.path.exists(
                            out_file_pop
                        ):
                            rMisc.rasterizeDataFrame(
                                inD, out_file, idField=field, templateRaster=template
                            )

                        # Multiply the rasterized data frame by the population layer
                        if not os.path.exists(out_file_pop):
                            combine_dhs_pop(
                                inP, rasterio.open(out_file), out_file_pop, factor=100
                            )
                        if os.path.exists(out_file):
                            os.remove(out_file)
                        misc.tPrint(f"{iso3}_{key}: {field}")
                        dhs_rasters[f"{key}_{field}"] = {
                            "raster_file": f"{key}_{field}_pop.tif",
                            "vars": ["SUM", "MEAN"],
                            "description": f"{key}_{field}",
                        }
                    except:
                        misc.tPrint(f"Error processing {key} - {field}")
    return dhs_rasters


def extract_data(inG, inG1, inG2, inL, inR):
    country_folder = os.path.join(output_folder, iso3)
    adm0_file = os.path.join(country_folder, "adm0.shp")
    adm1_file = os.path.join(country_folder, "adm1.shp")
    adm2_file = os.path.join(country_folder, "adm2.shp")
    lc_file = os.path.join(country_folder, "LC.tif")

    if not os.path.exists(country_folder):
        os.makedirs(country_folder)
    country_bounds = inG.loc[inG["ISO3"] == iso3].to_crs({"init": "epsg:4326"})
    if not os.path.exists(adm0_file):
        country_bounds.to_file(adm0_file)
    if not os.path.exists(adm1_file):
        try:
            country_adm1 = inG1.loc[inG1["ISO3"] == iso3].to_crs({"init": "epsg:4326"})
            country_adm1.to_file(adm1_file)
        except:
            misc.tPrint("%s Could not extract ADMIN 1" % iso3)
    if not os.path.exists(adm2_file):
        try:
            country_adm2 = inG2.loc[inG2["ISO3"] == iso3].to_crs({"init": "epsg:4326"})
            country_adm2.to_file(adm2_file)
        except:
            misc.tPrint("%s Could not extract ADMIN 2" % iso3)
    if not os.path.exists(lc_file):
        rMisc.clipRaster(inL, gpd.read_file(adm0_file), lc_file)

    calculate_vulnerability(iso3, country_folder, country_bounds, pop_folder, pop_files)
    misc.tPrint("***%s Calculated Vulnerability" % iso3)
    try:
        create_urban_data(iso3, country_folder, country_bounds, inR, calc_urban=False)
        misc.tPrint("***%s Calculated Urban Extents" % iso3)
    except:
        misc.tPrint("%s errored on HD clusters" % iso3)
        try:
            create_urban_data(
                iso3,
                country_folder,
                country_bounds,
                inR,
                calc_urban=True,
                calc_hd_urban=False,
            )
        except:
            misc.tPrint("%s errored on all clusters" % iso3)
    # extract_osm(country_bounds, country_folder)
    # misc.tPrint("***Extracted OSM")


def run_all(iso3, output_folder, dhs_files):
    country_folder = os.path.join(output_folder, iso3)
    # extract national bounds
    misc.tPrint("Processing %s" % iso3)
    # summarize DHS
    country_pop = os.path.join(country_folder, "WP_2020_1km.tif")
    dhs_rasters = summarize_DHS(country_pop, dhs_files, country_folder, iso3)

    # Run zonal stats
    cur_rasters = copy.deepcopy(hnp_categories)
    for key, values in cur_rasters.items():
        values["raster_file"] = os.path.join(country_folder, values["raster_file"])
        cur_rasters[key] = values

    cur_dhs = copy.deepcopy(dhs_rasters)
    for key, values in dhs_rasters.items():
        values["raster_file"] = os.path.join(country_folder, values["raster_file"])
        cur_dhs[key] = values

    all_shps = []
    for (
        root,
        dirs,
        files,
    ) in os.walk(country_folder):
        for f in files:
            if f[-4:] == ".shp" and "zonal" not in f:
                all_shps.append(os.path.join(root, f))

    run_zonal(all_shps, cur_rasters, out_suffix="_BASE", iso3=iso3)
    misc.tPrint("***%s Calculated Base Zonal" % iso3)
    run_zonal(all_shps, cur_dhs, out_suffix="_DHS", iso3=iso3)
    misc.tPrint("***%s Calculated DHS Zonal" % iso3)


def main():
    # define the input datasets
    global_bounds = "/home/public/Data/GLOBAL/ADMIN/Admin0_Polys.shp"
    global_adm1 = "/home/public/Data/GLOBAL/ADMIN/Admin1_Polys.shp"
    global_adm2 = "/home/public/Data/GLOBAL/ADMIN/Admin2_Polys.shp"
    pop_folder = (
        "/home/public/Data/GLOBAL/Population/WorldPop_PPP_2020/GLOBAL_1km_Demographics"
    )
    output_folder = "/home/wb411133/data/Projects/CoVID"
    population_raster = "/home/public/Data/GLOBAL/Population/WorldPop_PPP_2020/ppp_2020_1km_Aggregated.tif"
    lcRaster = "/home/public/Data/GLOBAL/LANDCOVER/GLOBCOVER/2015/ESACCI-LC-L4-LCCS-Map-300m-P1Y-2015-v2.0.7.tif"
    dhs_folder = "/home/public/Data/PROJECTS/CoVID/DHS"
    dhs_files = {}
    for root, dirs, files in os.walk(dhs_folder):
        for f in files:
            if f[-4:] == ".shp":
                dhs_files[f.replace(".shp", "")] = gpd.read_file(
                    os.path.join(os.path.join(root, f))
                )

    # Read in the global datasets
    pop_files = os.listdir(pop_folder)
    inG = gpd.read_file(global_bounds)
    inG1 = gpd.read_file(global_adm1)
    inG2 = gpd.read_file(global_adm2)
    inR = rasterio.open(population_raster)
    inL = rasterio.open(lcRaster)

    countries = [
        "ARG",
        "PAK",
        "ZAF",
        "COL",
        "ZWE",
        "MNG",
        "SLE",
        "CPV",
        "KEN",
        "GHA",
        "AFG",
        "YEM",
        "ECU",
        "PRY",
        "MRT",
        "MDV",
        "KGZ",
        "HTI",
        "DJI",
        "KHM",
        "TJK",
        "GMB",
        "LKA",
        "SEN",
        "STP",
        "SLV",
        "VEN",
        "MLI",
        "RWA",
        "BOL",
        "TZA",
        "MAR",
        "IND",
        "IDN",
        "SDN",
        "AGO",
        "BEN",
        "BWA",
        "BFA",
        "BDI",
        "CMR",
        "CAF",
        "TCD",
        "COM",
        "COG",
        "CIV",
        "COD",
        "SSD",
        "ERI",
        "ETH",
        "GAB",
        "GNB",
        "GIN",
        "LSO",
        "LBR",
        "MDG",
        "MWI",
        "MUS",
        "MOZ",
        "NAM",
        "NER",
        "NGA",
        "SYC",
        "SOM",
        "SWZ",
        "TGO",
        "UGA",
        "ZMB",
        "LCA",
        "PHL",
        "GTM",
        "BGD",
        "BRA",
        "MEX",
        "EGY",
        "UKR",
        "PER",
        "LAO",
        "PSE",
        "NPL",
        "PNG",
        "DZA",
        "BLR",
        "BTN",
        "BIH",
        "NIC",
        "FJI",
        "GEO",
        "HND",
        "JOR",
        "MHL",
        "MDA",
        "MMR",
        "MKD",
        "PAN",
        "WSM",
        "SLB",
        "TUN",
        "TUR",
        "URY",
        "UZB",
        "ALB",
        "HRV",
        "IRN",
        "SRB",
        "TTO",
        "ATG",
        "CHN",
        "IRQ",
    ]
    countries = ["EGY", "PHL", "BGD"]
    countries = set(countries)
    nCountries = len(countries)
    idx = 0
    all_commands = []
    for iso3 in countries:
        all_commands.append([iso3, output_folder, dhs_files])

    with Pool(round(multiprocessing.cpu_count() * 0.8)) as p:
        res = p.starmap(run_all, all_commands)


if __name__ == "__main__":
    main()



import sys
import os
import inspect
import rasterio

import numpy as np

from GOSTrocks.misc import tPrint

curPath = os.path.realpath(
    os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])
)
if curPath not in sys.path:
    sys.path.append(curPath)


def combine_ghsl_annual(
    ghsl_files, built_thresh=0.1, ghsl_files_labels=[], out_file=""
):
    """_summary_

    :param ghsl_files: list of ghsl annual files to process
    :type ghsl_files: list of strings (paths to ghsl files)
    :param built_thresh: minimum percetn built to be considered bult, defaults to 0.1 which is 10%
    :type built_thresh: float, optional
    :param ghsl_files_labels: list of numbers to define values in output raster, defaults to [] which means numbers will be extracted from the files.
    :type ghsl_files_labels: list of ints
    :param out_file: location to write output integer file, defaults to '' which does not write anything
    :type out_file: str, optional
    :returns: list of ghsl values and rasterio profile
    :rtype: list of [numpy array, dictionary]
    """

    # open all the ghsl files, extract data and labels
    ghsl_rasters = []
    ghsl_years = []
    idx = 0
    for ghsl_file in ghsl_files:
        cur_r = rasterio.open(ghsl_file)
        out_meta = cur_r.profile.copy()
        cur_d = cur_r.read()[0, :, :]
        cur_d[cur_d == cur_r.profile["nodata"]] = 0
        if len(ghsl_files_labels) > 0:
            cur_year = ghsl_files_labels[idx]
        cur_year = ghsl_file.split("_")[-7][1:]

        # Convert built area to dataset with single value of the current year
        cur_d = ((cur_d >= built_thresh) * int(cur_year)).astype(float)
        cur_d[cur_d == 0] = np.nan

        ghsl_rasters.append(cur_d)
        ghsl_years.append(cur_year)

        tPrint(f"*** {idx} completed {cur_year}")
        idx += 1

    # stack the ghsl files
    all_ghsl = np.dstack(ghsl_rasters)
    ghsl_final = np.nanmin(all_ghsl, axis=2)

    ghsl_int = ghsl_final.astype(int)
    ghsl_int[ghsl_int < 0] = int(out_meta["nodata"])

    # write output
    if out_file != "":
        with rasterio.open(out_file, "w", **out_meta) as out_r:
            out_r.write_band(1, ghsl_int)

    return [ghsl_int, out_meta]



import os
import rasterio.mask

import pandas as pd

import GOSTnets.load_osm as losm

from . import osm_extractor as osm
from . import rasterMisc as rMisc


def extract_rai_network(focal_osm, epsg=3857, rai_buffer=2000):
    """extract road network from OSM PBF

    INPUTS
    focal_osm [string] - path to osm pbf to be used for calculations
    epsg [int] - epsg code for metres-based measurement codes
    rai_buffer [number]

    RETURNS
    [geopandas data frame]
    """
    G_loader = losm.OSM_to_network(focal_osm)
    G_loader.generateRoadsGDF()
    roadsGPD = G_loader.roadsGPD
    roadsGPD.set_geometry("Wkt", inplace=True)
    roadsGPD = roadsGPD.to_crs({"init": "epsg:%s" % epsg})
    roadsGPD_rai = roadsGPD.copy()
    roadsGPD_rai["Wkt"] = roadsGPD_rai["Wkt"].apply(lambda x: x.buffer(rai_buffer))

    roadsGPD_rai["OSMLR"] = roadsGPD_rai["infra_type"].map(osm.OSMLR_Classes)

    def get_num(x):
        try:
            return int(x)
        except:
            return 5

    roadsGPD_rai["OSMLR_num"] = roadsGPD_rai["OSMLR"].apply(
        lambda x: get_num(str(x)[-1])
    )

    return roadsGPD_rai


def calculate_rai(out_bounds, out_bounds_id, wp_data, roadsGPD_rai, rai_folder):
    """Calculate RAI

    INPUT
    out_bounds [geopandas dataframe] - admin boundaries within which to summarize RAI
    out_bounds_id [string] - column name in out_bounds containing INDETIFIER
    wp_data [rasterio raster] - gridded population dataset from which to calculate RAI
    roadsGPD_rai [geopandas data frame] - generated from function .extract_rai_network()

    RETURNS
    [pandas dataframe]
    """
    actual_pop = rMisc.zonalStats(out_bounds, wp_data, minVal=0)
    actual_pop = pd.DataFrame(actual_pop, columns=["SUM", "MIN", "MAX", "STDEV"])
    # Generate OSMLR level road maps
    all_res = {}
    for road_count in [1, 2, 3, 4]:
        rai_pop = os.path.join(rai_folder, "WP_RAI_%s.tif" % road_count)
        # Generate pop map within RAI pop
        if not os.path.exists(rai_pop):
            cur_roads = roadsGPD_rai.loc[roadsGPD_rai["OSMLR_num"] <= road_count]
            roads_mask = rasterio.mask.mask(
                wp_data, [cur_roads.unary_union], invert=False
            )
            with rasterio.open(rai_pop, "w", **wp_data.meta) as out:
                out.write(roads_mask[0])
        rai_pop = rMisc.zonalStats(out_bounds, rai_pop, minVal=0)

        rai_pop = pd.DataFrame(rai_pop, columns=["SUM", "MIN", "MAX", "STDEV"])
        rai_pop["ID"] = out_bounds[out_bounds_id]
        rai_pop = rai_pop[["SUM", "ID"]]
        rai_pop.columns = ["RAI_POP_%s" % road_count, "ID"]
        all_res[road_count] = rai_pop

    final = all_res[1]
    final["RAI_POP_2"] = all_res[2]["RAI_POP_2"]
    final["RAI_POP_3"] = all_res[3]["RAI_POP_3"]
    final["RAI_POP_4"] = all_res[4]["RAI_POP_4"]
    final["POP"] = actual_pop["SUM"]
    return final



import sys
import os
import inspect
import rasterio

import pandas as pd
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt

curPath = os.path.realpath(
    os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])
)
if curPath not in sys.path:
    sys.path.append(curPath)

from dataMisc import aws_search_ntl
from misc import tPrint
import rasterMisc as rMisc


def read_raster_box(curRaster, geometry, bandNum=1):
    # get pixel coordinates of the geometry's bounding box
    ul = curRaster.index(*geometry.bounds[0:2])
    lr = curRaster.index(*geometry.bounds[2:4])
    # read the subset of the data into a numpy array
    window = ((float(lr[0]), float(ul[0] + 1)), (float(ul[1]), float(lr[1] + 1)))
    data = curRaster.read(bandNum, window=window)
    return data


def calc_annual(df, extent, agg_method="MEAN"):
    """Combine monthly nighttime lights images into an annual composite

    :param df: data frame of images with three columns: YEAR, MONTH, PATH
    :type df: pandas.DataFrame
    :param extent: area to extract imagery from
    :type extent: shapely.Polygon
    """
    all_layers = df["PATH"].apply(lambda x: read_raster_box(rasterio.open(x), extent))
    all_vals = np.dstack(all_layers)
    if agg_method == "MEAN":
        final_vals = np.nanmean(all_vals, axis=2)

    return final_vals


def generate_annual_composites(aoi, agg_method="MEAN", sel_files=[], out_folder=""):
    """_summary_

    :param aoi: geopandas polygonal dataframe to use for clip clip extent based on crop param
    :type aoi: geopandas.GeoDataFrame
    :param method: How to aggregate monthly nighttime lights layers into annual layers, defaults to MEAN
    :type method: str, optional
    :param sel_files: list of ntl files to process, defaults to [], which will use gostrocks.dataMisc.aws_search_ntl to find all variables
    :type sel_files: list, optional
    """
    if len(sel_files) == 0:
        sel_files = aws_search_ntl()
    yr_month = [x.split("_")[1] for x in sel_files]
    yr = [x[:4] for x in yr_month]
    information = pd.DataFrame(
        [yr, yr_month, sel_files], index=["YEAR", "MONTH", "PATH"]
    ).transpose()
    annual_vals = information.groupby("YEAR").apply(lambda x: calc_annual(x, aoi))

    # Write the files to output
    if out_folder != "":
        if not os.path.exists(out_folder):
            os.makedirs(out_folder)
        out_meta = rasterio.open(information["PATH"].iloc[0]).profile.copy()
        for label, res in annual_vals.items():
            out_meta.update(
                width=res.shape[0],
                height=res.shape[1],
                transform=rasterio.transform.from_bounds(
                    *aoi.bounds, res.shape[0], res.shape[1]
                ),
            )
            out_file = os.path.join(out_folder, f"VIIRS_{label}_annual.tif")
            with rasterio.open(out_file, "w", **out_meta) as out_r:
                out_r.write_band(1, res)
    return annual_vals


def map_viirs(
    cur_file,
    out_file="",
    class_bins=[-10, 0.5, 1, 2, 3, 5, 10, 15, 20, 30, 40, 50],
    text_x=0,
    text_y=5,
    dpi=100,
):
    """Map VIIRS nighttime lights imagery, optionally create output image

    :param cur_file: path to input geotiff
    :type cur_file: string
    :param out_file: path to create output image, defaults to '' which does not create a file
    :type out_file: str, optional
    :param class_bins: breaks for applying colour ramp, defaults to [-10,0.5,1,2,3,5,10,15,20,30,40,50]
    :type class_bins: list, optional
    :param text_x: position on map to position year text (left to right), defaults to 0
    :type text_x: int, optional
    :param text_y: position on map to position year text (top to bottom), defaults to 5
    :type text_y: int, optional
    :param dpi: dotes per inch for output image, defaults to 100
    :type dpi: int, optional
    """
    # extract the year from the file name
    year = cur_file.split("_")[-1][:4]

    # Open the VIIRS data and reclassify
    inR = rasterio.open(cur_file)
    inD = inR.read()
    inC = xr.apply_ufunc(np.digitize, inD, class_bins)

    # Plot the figure, remove grid and ticks
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.grid(False)
    ax.set_xticks([])
    ax.set_yticks([])

    ### TODO: add the year to the map, may need to experiment with the location depend on geography
    ax.text(text_x, text_y, year, fontsize=40, color="white")

    # plt.margins(0,0)
    if out_file != "":
        # plt.imsave(out_file, inC[0,:,:], cmap=plt.get_cmap('magma'))
        plt.imshow(inC[0, :, :], cmap=plt.get_cmap("magma"))
        fig.savefig(out_file, dpi=dpi, bbox_inches="tight", pad_inches=0)
    else:
        # https://matplotlib.org/stable/tutorials/colors/colormaps.html
        plt.imshow(inC[0, :, :], cmap=plt.get_cmap("magma"))


def run_zonal(inD, ntl_files=[], minval=0.1, verbose=False, calc_sd=True):
    """Run zonal statistics against a series of nighttime lights files

    :param inD: input geopandas dataframe in which to summarize results
    :type inD: gpd.GeoDataFrames
    :param ntl_files: list of ntl files to summarize, defaults to [] which will search for all files in the s3 bucket using datMisc.aws_search_ntl()
    :type ntl_files: list, optional
    :param minval: Minimum value to summarize in nighttime lights, defaults to 0.1 which means all values below this become 0
    :type minval: float, optional
    :param verbose: print additional information, defaults to False
    :type verbose: bool, optional
    :param calc_sd: _description_, defaults to True
    :type calc_sd: bool, optional
    """

    """ run zonal stats on all ntl files
    INPUT
        inD [geopandas dataframe]

    RETURNS
        pandas dataframe
    """
    if len(ntl_files) == 0:
        ntl_files = aws_search_ntl()

    for ntl_file in ntl_files:
        name = ntl_file.split("/")[-1].split("_")[2][:8]
        if verbose:
            tPrint(name)
        inR = rasterio.open(ntl_file)
        ntl_res = rMisc.zonalStats(inD, inR, minVal=minval, calc_sd=calc_sd)
        out_cols = ["SUM", "MIN", "MAX", "MEAN"]
        if calc_sd:
            out_cols.append("SD")
        ntl_df = pd.DataFrame(ntl_res, columns=out_cols)
        inD[f"ntl_{name}_SUM"] = ntl_df["SUM"]
    return inD



import subprocess
import argparse
import logging

import geopandas as gpd
import osmnx as ox
import pandas as pd

from shapely.geometry import box
from GOSTrocks import misc

# Highway features are reclassified to 4 OSMLR classes for simplification and standardization
#   https://mapzen.com/blog/osmlr-2nd-technical-preview/
OSMLR_Classes = {
    "motorway": "OSMLR level 1",
    "motorway_link": "OSMLR level 1",
    "trunk": "OSMLR level 1",
    "trunk_link": "OSMLR level 1",
    "primary": "OSMLR level 1",
    "primary_link": "OSMLR level 1",
    "secondary": "OSMLR level 2",
    "secondary_link": "OSMLR level 2",
    "tertiary": "OSMLR level 2",
    "tertiary_link": "OSMLR level 2",
    "unclassified": "OSMLR level 3",
    "unclassified_link": "OSMLR level 3",
    "residential": "OSMLR level 3",
    "residential_link": "OSMLR level 3",
    "track": "OSMLR level 4",
    "service": "OSMLR level 4",
}


class osmExtraction(object):
    """
    Download and installation instructions and basic usage can be found here
    https://wiki.openstreetmap.org/wiki/Osmosis
    """

    def __init__(
        self,
        osmosisCmd=r"C:\WBG\Anaconda\Osmosis\bin\osmosis",
        tempFile=r"C:\WBG\Anaconda\Osmosis\tempExecution.bat",
    ):
        """
        osmosisCmd is the path to the file osmosis in the bin folder of the downloaded
            stable version of osmosis (link above)
        """
        self.osmosisCommand = osmosisCmd
        self.tempFile = tempFile

    def extractAmmenities(
        self, inPbf, outFile, amenityList=["amenity.school"], bounds=[], execute=False
    ):
        """Read input osmpbf, extract all buildings and spit out to shapefile
        INPUT
        inPbf [string] - path to input pbf
        outFile [string] - path to output shapefile

        OPTIONAL
        amenity [string] - definition of amenity to extract - defaults to schools
        bounds [list of numbers??] - bounds of area to extract, if not defined, covers entire input area
        """
        if len(amenityList) > 1:
            amenity = ",".join(amenityList)
        else:
            amenity = amenityList[0]

        baseCommand = (
            r'{osmCommand} --read-pbf {inPbf} --nkv keyValueList="{amenity}"'.format(
                osmCommand=self.osmosisCommand, inPbf=inPbf, amenity=amenity
            )
        )
        if len(bounds) > 0:
            baseCommand = "{baseCommand} --bounding-box top={top} left={left} bottom={bottom} right={right}".format(
                baseCommand=baseCommand,
                top=bounds[3],
                left=bounds[0],
                bottom=bounds[1],
                right=bounds[2],
            )

        baseCommand = "{baseCommand}  --write-pbf {outPbf}".format(
            baseCommand=baseCommand, outPbf=outFile
        )
        if not execute:
            return baseCommand
        else:
            with open(self.tempFile, "w") as outFile:
                outFile.write(baseCommand)
            subprocess.call([self.tempFile], shell=True)

    def extractBuildings(self, inPbf, outFile, bounds=[], execute=True):
        """Read input osmpbf, extract all buildings and spit out to shapefile
        INPUT
        inPbf [string] - path to input pbf
        outFile [string] - path to output shapefile
        """
        baseCommand = (
            r"{osmCommand} --read-pbf {inPbf} --tf accept-ways building=*".format(
                osmCommand=self.osmosisCommand, inPbf=inPbf
            )
        )
        if len(bounds) > 0:
            baseCommand = "{baseCommand} --bounding-box top={top} left={left} bottom={bottom} right={right}".format(
                baseCommand=baseCommand,
                top=bounds[3],
                left=bounds[0],
                bottom=bounds[1],
                right=bounds[2],
            )
        baseCommand = "{baseCommand}  --write-pbf {outPbf}".format(
            baseCommand=baseCommand, outPbf=outFile
        )
        if not execute:
            return baseCommand
        else:
            with open(self.tempFile, "w") as outFile:
                outFile.write(baseCommand)
            subprocess.call([self.tempFile], shell=True)

    def extractBoundingBox(self, inPbf, inShp, outPbf, execute=True):
        """Extract the inPbf based on the bounding box of the input shapefile
        INPUT
        inPbf [string] - path to input pbf
        inShp [string or geopandas object] - path to aoi shapefile
        """
        if type(inShp) == str:
            inD = gpd.read_file(inShp)
        elif type(inShp) == gpd.GeoDataFrame:
            inD = inShp
        else:
            raise (ValueError("inShp needs to be a string or a geopandas object"))
        if inD.crs != {"init": "epsg:4326"}:
            inD = inD.to_crs({"init": "epsg:4326"})
        baseCommand = r"{osmCommand} --read-pbf {inPbf} --bounding-box top={top} left={left} bottom={bottom} right={right} --write-pbf {outPbf}".format(
            osmCommand=self.osmosisCommand,
            inPbf=inPbf,
            top=float(inD.bounds.maxy),
            left=float(inD.bounds.minx),
            bottom=float(inD.bounds.miny),
            right=float(inD.bounds.maxx),
            outPbf=outPbf,
        )
        if not execute:
            return baseCommand
        else:
            with open(self.tempFile, "w") as outFile:
                outFile.write(baseCommand)
            subprocess.call([self.tempFile], shell=True)

    def extractHighways(
        self, inPbf, outOSM, values=[1, 2, 3, 4], bounds=[], execute=True
    ):
        """Extract highways (all roads) from input osm pbf. Can limit definition of which roads to extract
            by defining the OSMLT class. see osmMisc.OSMLR_Classes for definition of classes

        INPUT
            inPbf [string]  - path to input osm pbf file
            outOSM [string] - path to output osm pbf file
            values [list of int] [optional][default = [1,2,3,4] - OSMLR values to extract
            bounds [list of coordinates] - boundary to extract
            execute [boolean] [default = True] - if set to false, command will return the osmosis command, and not execute it
        """
        highwayVals = []
        for key, value in OSMLR_Classes.items():
            try:
                if int(value.split(" ")[2]) in values:
                    # highwayVals.append("highway=%s" % key)
                    highwayVals.append(key)
            except:
                pass
        allCommands = ",".join(highwayVals)
        baseCommand = r"{osmCmd} --read-pbf {inPbf} --tf accept-ways highway={highwayCommand} --used-node".format(
            osmCmd=self.osmosisCommand, inPbf=inPbf, highwayCommand=allCommands
        )

        if len(bounds) > 0:
            baseCommand = "{baseCommand} --bounding-box top={top} left={left} bottom={bottom} right={right}".format(
                baseCommand=baseCommand,
                top=bounds[3],
                left=bounds[0],
                bottom=bounds[1],
                right=bounds[2],
            )

        baseCommand = "{baseCommand} --write-pbf {outPbf}".format(
            baseCommand=baseCommand, outPbf=outOSM
        )
        if not execute:
            return baseCommand
        else:
            with open(self.tempFile, "w") as outFile:
                outFile.write(baseCommand)
            subprocess.call([self.tempFile], shell=True)


def summarizeOSM(grid, verbose=True, roadsOnly=False):
    """Summarizes OSM road length within each feature in the input grid

    ---variables---
    grid [GeoDataframe] - each feature will have the length of all roads summarized

    --- To Do ---
    1. The length projection is web mercator - this is not great, and should instead be projected to UTM
    """
    WGS_84 = {"init": "epsg:4326"}
    WEB_MERCATOR = {"init": "epsg:3857"}
    # Extract OSM within the bounding box of the input grid
    if grid.crs != WGS_84:
        grid = grid.to_crs(WGS_84)
    bbox = box(
        grid.bounds.minx.min(),
        grid.bounds.miny.min(),
        grid.bounds.maxx.max(),
        grid.bounds.maxy.max(),
    )
    G = ox.graph_from_polygon(bbox, network_type="drive_service")
    # Limit the OSM grid to important columns, ignore nodes
    nodes, roads = ox.save_load.graph_to_gdfs(G, edges=True)
    if roadsOnly:
        return roads
    roads = roads[["geometry", "highway", "osmid"]]
    roads["length"] = roads.length
    roads["highway"] = roads.highway.astype(str)
    roads["OSMLR"] = roads.highway.map(OSMLR_Classes)
    FID_list, OSMLR1_list, OSMLR2_list, OSMLR3_list, OSMLRt_list = [], [], [], [], []
    cnt = 0
    verbose = False
    # Loop through all of the input features to summarize OSM
    for idx, obj in grid.iterrows():
        if idx % 50 == 0 and verbose:
            print("%s of %s" % (cnt, len(grid.shape[0])))
        roads2 = roads.copy()
        # Find roads that intersect current geometry
        roads2 = roads2[roads2.intersects(obj.geometry)]
        # Find the intersection of the current geometry and the intersecting roads
        roads2["intersecting"] = roads2.geometry.intersection(obj.geometry)
        roads2 = roads2.set_geometry("intersecting")
        roads2["intersecting_length"] = roads2.length
        # Project the roads to a metre-based CRS
        roads2 = roads2.to_crs(WEB_MERCATOR)
        roads2["intersecting_length"] = roads2.length
        FID_list.append(idx)
        # Summarize total lenght of OSMLR classes
        OSMLR1_list.append(
            roads2.loc[roads2.OSMLR == "OSMLR level 1"].intersecting_length.sum()
        )
        OSMLR2_list.append(
            roads2.loc[roads2.OSMLR == "OSMLR level 2"].intersecting_length.sum()
        )
        OSMLR3_list.append(
            roads2.loc[roads2.OSMLR == "OSMLR level 3"].intersecting_length.sum()
        )
        OSMLRt_list.append(
            roads2.loc[roads2.OSMLR == "OSMLR track"].intersecting_length.sum()
        )
        cnt = cnt + 1

    results = pd.DataFrame(
        {
            "FID": FID_list,
            "OSMLR level 1": OSMLR1_list,
            "OSMLR level 2": OSMLR2_list,
            "OSMLR level 3": OSMLR3_list,
            "OSMLR track": OSMLRt_list,
        }
    )
    results["totalOSM"] = (
        results["OSMLR level 1"]
        + results["OSMLR level 2"]
        + results["OSMLR level 3"]
        + results["OSMLR track"]
    )
    return results


def convertOSMPBF_DataFrame(inOSM, layer):
    """Convert an input OSM PBF to a geopandas dataframe

    INPUT
    inOSM [string] - path to OSM pbf to convert
    layer [string] - data layer to extract. Select from lines, points, multipolygons, multilinestrings

    RETURNS
    [geopandas data frame]
    """
    import geojson
    import json
    from osgeo import ogr
    from shapely.geometry import shape

    driver = ogr.GetDriverByName("OSM")
    data = driver.Open(inOSM)
    layer = data.GetLayer(layer)
    features = [x for x in layer]

    def loadOSMjson(x):
        """convert the OSM JSON to something concumable as geopandas"""
        geom = shape(geojson.loads(json.dumps(x["geometry"])))
        x["geometry"] = geom
        for key, value in x["properties"].items():
            x[key] = value
        try:
            splitTags = x["other_tags"].split(",")
            for tag in splitTags:
                tSplit = tag.split("=>")
                x[tSplit[0].replace('"', "")] = tSplit[1].replace('"', "")
        except:
            pass
        x.pop("properties", None)
        x.pop("other_tags", None)
        return x

    allFeats = [loadOSMjson(feat.ExportToJson(as_object=True)) for feat in features]
    inR = pd.DataFrame(allFeats)
    inR = gpd.GeoDataFrame(inR, geometry="geometry", crs=4326)
    try:
        inR = misc.project_UTM(inR)
    except:
        inR = inR.to_crs(3857)
    inR["Length"] = inR["geometry"].apply(lambda x: x.length)
    inR = inR.to_crs(4326)
    return inR


if __name__ == "__main__":
    exampleText = """
    #Extract OSMLR level 1 and 2 roads from pbf
    python osmMisc.py -i Q:\AFRICA\MRT\INFRA\mauritania-latest_20190103.osm.pbf -o Q:\AFRICA\MRT\INFRA\mauritania-latest_20190103_OSMLR12.osm.pbf -OSMLR 1 2 -OSMExtract

    #Extract OSM PBF for bounding box
    python osmMisc.py -i Q:\AFRICA\LBR\INFRA\liberia-latest.osm.pbf -o Q:\AFRICA\LBR\INFRA\liberia-latest_Monrovia.osm.pbf -AOI_file "Q:\AFRICA\LBR\Projects\Monrovia_Imagery\Greater Monrovia admin boundary\Greater_Monrovia_admin_boundary.shp" -BBExtract

    #Extract buildings from OSM PBF
    python osmMisc.py -i Q:\AFRICA\LBR\INFRA\liberia-latest_Monrovia.osm.pbf -o Q:\AFRICA\LBR\INFRA\liberia-latest_Monrovia_Buildings.osm.pbf -BuildingExtract
    """
    parser = argparse.ArgumentParser(
        description="Generate urban metrics based on a defined grid"
    )

    parser.add_argument("-i", dest="INPUT", action="store", help="Input .osm.pbf")
    parser.add_argument("-o", dest="OUTPUT", action="store", help="Output .osm.pbf")
    parser.add_argument(
        "-OSMExtract",
        dest="EXTRACT",
        action="store_true",
        help="Perform OSMExtract on OSM PBF",
    )
    parser.add_argument(
        "-OSMLR",
        dest="LEVELS",
        action="store",
        help="OSMLR levels to extract from the input pbf",
        nargs="+",
    )
    parser.add_argument(
        "-BBExtract",
        dest="BBEXTRACT",
        action="store_true",
        help="Perform OSMExtract on OSM PBF",
    )
    parser.add_argument(
        "-BuildingExtract",
        dest="BUILDINGEXTRACT",
        action="store_true",
        help="Perform OSMExtract on OSM PBF",
    )
    parser.add_argument(
        "-AOI_file", dest="AOI", action="store", help="AOI of study area"
    )

    args = parser.parse_args()
    print(args)
    # Set logging information
    logging.basicConfig(
        format="%(asctime)s:%(levelname)s:  %(message)s", level=logging.INFO
    )
    xx = osmExtraction()
    if args.EXTRACT:
        y = xx.extractOSMLRlvl1(
            args.INPUT, args.OUTPUT, values=[int(x) for x in args.LEVELS], execute=True
        )
        print(y)
    if args.BBEXTRACT:
        y = xx.extractBoundingBox(args.INPUT, args.AOI, args.OUTPUT)
        print(y)
    if args.BUILDINGEXTRACT:
        y = xx.extractBuildings(args.INPUT, args.OUTPUT)
        print(y)



import sys
import os
import inspect
import json
import rasterio
import pyproj

import pandas as pd
import geopandas as gpd
import numpy as np

from shapely.geometry import box, shape
from shapely import wkt
from affine import Affine
from rasterio import features
from rasterio.mask import mask
from rasterio.features import rasterize, MergeAlg
from rasterio.warp import reproject, Resampling, calculate_default_transform
from rasterio.merge import merge
from rasterio.io import MemoryFile
from contextlib import contextmanager

curPath = os.path.realpath(
    os.path.abspath(os.path.split(inspect.getfile(inspect.currentframe()))[0])
)
if curPath not in sys.path:
    sys.path.append(curPath)

from misc import tPrint


def merge_rasters(
    in_rasters, merge_method="first", dtype="", out_file="", boolean_gt_0=False
):
    """Merge a list of rasters into a single raster file

    Args:
        in_rasters (list of rasters): List of rasters to process as string paths
        merge_method (str, optional): merge method. Defaults to 'first'.
        dtype (str, optional): dtype to convert final raster to. Defaults to '', which maintains input raster type
        out_file (str, optional): Path to create output raster. Defaults to '', which creates no raster
        boolean_gt_0 (str, optional): If true, converts merged result to binary 0, 1 of values greater than 0, Defaults to False
    """

    opened_tiffs = [rasterio.open(x) for x in in_rasters]
    merged, out_transform = merge(opened_tiffs, method=merge_method)
    if boolean_gt_0:
        merged = (merged > 0) * 1
        dtype = "uint8"
    if dtype != "":
        merged = merged.astype(dtype)

    # Create a new raster file with the merged data
    metadata = opened_tiffs[0].meta.copy()
    metadata.update(
        {
            "height": merged.shape[1],
            "width": merged.shape[2],
            "transform": out_transform,
            "dtype": "uint8",
        }
    )
    if out_file != "":
        with rasterio.open(out_file, "w", **metadata) as dst:
            dst.write(merged)
    return (merged, metadata)


@contextmanager
def create_rasterio_inmemory(src, curData):
    """Create a rasterio object in memory from a numpy array

    :param src: data dictionary describing the rasterio template i.e. - rasterio.open().profile
    :type src: rasterio metadata dictionary
    :param curData: numpy array from which to create rasterio object
    :type curData: numpy array
    """
    with MemoryFile() as memFile:
        with memFile.open(**src) as dataset:
            try:
                dataset.write(curData)
            except:
                dataset.write_band(1, curData)
            del curData

        with memFile.open() as dataset:
            yield dataset


def vectorize_raster(
    inR, bad_vals=[]
):  # TODO out_file='', smooth=False, smooth_window=3, bad_vals=None):
    """convert input raster data to a geodatframe

    :param inR: input raster data to vectorize
    :type inR: rasterio.datasetReader
    :param bad_vals: list of values to ignore in conversion
    :type bad_vals: list of int
    :returns: geopandas data frame of [idx, value (from raster), and geometry]
    :rtype: geopandas.GeoDataFrame
    """

    data = inR.read()
    ## TODO add smoothing option
    # if smooth:
    ## TODO add bad value filtering
    idx = 0
    all_vals = []
    for cShape, value in features.shapes(data, transform=inR.transform):
        if value not in bad_vals:
            all_vals.append([idx, value, shape(cShape)])
            # shape(geojson.loads(json.dumps(cShape)))
            idx += 1

    return gpd.GeoDataFrame(
        all_vals, columns=["idx", "value", "geometry"], geometry="geometry", crs=inR.crs
    )


def project_raster(srcRst, dstCrs, output_raster=""):
    """project raster to destination crs

    Args:
        srcRst (_type_): _description_
        dstCrs (_type_): _description_
        output_raster (_type_): _description_
    """
    transform, width, height = calculate_default_transform(
        srcRst.crs, dstCrs, srcRst.width, srcRst.height, *srcRst.bounds
    )
    kwargs = srcRst.meta.copy()
    kwargs.update(
        {"crs": dstCrs, "transform": transform, "width": width, "height": height}
    )

    # open destination raster
    dstRst = np.zeros([kwargs["count"], width, height], kwargs["dtype"])

    # reproject and save raster band data
    for i in range(1, srcRst.count + 1):
        reproject(
            source=rasterio.band(srcRst, i),
            destination=dstRst,
            # src_transform=srcRst.transform,
            src_crs=srcRst.crs,
            # dst_transform=transform,
            dst_crs=dstCrs,
            resampling=Resampling.nearest,
        )

    if output_raster != "":
        with rasterio.open(output_raster, "w", *kwargs) as out_raster:
            out_raster.write(dstRst)

    return [dstRst, kwargs]


def clipRaster(inR, inD, outFile=None, crop=True):
    """Clip input raster

    :param inR: rasterio object to clip
    :type inR: rasterio.DatasetReader
    :param inD: geopandas polygonal dataframe to use for clip clip extent based on crop param
    :type inD: geopandas.GeoDataFrame
    :param outFile: string path to write output raster, default is '' which writes nothing
    :type outFile: string
    :param crop: determine wether to clip based on bounding box (False) or unary_union (True). Default is True
    :type crop: Boolean
    :return: array of [numpy array of data, and rasterio metadata]
    :rtype: array
    """
    if inD.crs != inR.crs:
        inD = inD.to_crs(inR.crs)
        inD = inD.buffer(0)
    out_meta = inR.meta.copy()

    def getFeatures(gdf):
        # Function to parse features from GeoDataFrame in such a manner that rasterio wants them
        return [json.loads(gdf.to_json())["features"][0]["geometry"]]

    if crop:
        tD = gpd.GeoDataFrame([[1]], geometry=[inD.unary_union])
    else:
        tD = gpd.GeoDataFrame([[1]], geometry=[box(*inD.total_bounds)])

    coords = getFeatures(tD)
    out_img, out_transform = mask(inR, shapes=coords, crop=True, all_touched=True)
    out_meta.update(
        {
            "driver": "GTiff",
            "height": out_img.shape[1],
            "width": out_img.shape[2],
            "transform": out_transform,
        }
    )
    if outFile:
        with rasterio.open(outFile, "w", **out_meta) as dest:
            dest.write(out_img)
    return [out_img, out_meta]


def rasterizeDataFrame(
    inD,
    outFile,
    idField="",
    templateRaster="",
    templateMeta="",
    nCells=0,
    res=0,
    mergeAlg="REPLACE",
    re_proj=False,
    nodata=np.nan,
):
    """Convert input geopandas dataframe into a raster file

    :param inD: input data frame to rasterize
    :type inD: gpd.GeoDataFrame
    :param outFile: output file to create from rasterized dataframe, defaults to '' which creates no file
    :type outFile: string
    :param idField: field in inD to rasterize, defaults to '' which sets everything to 1
    :type idField: str, optional
    :param templateRaster: raster upon which to base raster creation, defaults to ''. If no template is provided, nCells or res need to be defined
    :type templateRaster: str, optional
    :param templateMeta: raster metadata used to create output raster, defaults to ''. If no template is provided, nCells or res need to be defined
    :type templateMeta: str, optional
    :param nCells: number of cells in width and height, defaults to 0
    :type nCells: int, optional
    :param res: resolution of output raster in units of the crs, defaults to 0
    :type res: int, optional
    :param mergeAlg: Method of aggregating overlapping features, defaults to "REPLACE"; options are "REPLACE" or "ADD"
    :type mergeAlg: str, optional
    :param re_proj: option to reproject inD to templateRaster if CRS do not match, defaults to False
    :type re_proj: bool, optional
    :return:  dict of metadata used to create output raster and burned raster values
    :rtype: dict of {'meta':new raster metadata, 'vals': values in new raster}
    """
    ###Parameter checking
    if nCells <= 0 and res <= 0 and templateRaster == "" and templateMeta == "":
        raise (ValueError("Must define one of nCells or res"))
    if nCells > 0 and res > 0 and templateRaster == "" and templateMeta == "":
        raise (ValueError("Cannot define both nCells and res"))

    # Set VALUE field equal to idField
    inD["VALUE"] = 1
    inD["VALUE"] = inD["VALUE"].astype("int16")
    if idField != "":
        inD["VALUE"] = inD[idField]

    # set merge algorithm for overlapping features
    if mergeAlg == "REPLACE":
        mAlg = MergeAlg.replace
    elif mergeAlg == "ADD":
        mAlg = MergeAlg.add
    else:
        raise (ValueError("MergeAlg must be one of REPLACE or ADD"))

    if templateRaster != "":
        inR = rasterio.open(templateRaster)
        cMeta = inR.profile.copy()
        cMeta.update(count=1)
        nTransform = cMeta["transform"]
        if inD.crs != inR.crs:
            if not re_proj:
                raise (
                    ValueError(
                        "input CRS do not match: inD - %s, templateRaster - %s"
                        % (inD.crs, inR.crs)
                    )
                )
            inD = inD.to_crs(inR.crs)
    elif templateMeta != "":
        cMeta = templateMeta
        nTransform = cMeta["transform"]
        if inD.crs != cMeta["crs"]:
            if not re_proj:
                raise (
                    ValueError(
                        "input CRS do not match: inD - %s, templateRaster - %s"
                        % (inD.crs, inR.crs)
                    )
                )
            inD = inD.to_crs(cMeta["crs"])
    else:
        bounds = inD.total_bounds
        if nCells > 0:
            height = nCells
            width = nCells
        if res > 0:
            height = int(round((bounds[3] - bounds[1]) / res))
            width = int(round((bounds[2] - bounds[0]) / res))

        b = inD.total_bounds
        nTransform = rasterio.transform.from_bounds(
            b[0], b[1], b[2], b[3], width, height
        )
        if inD.crs.__class__ == pyproj.crs.crs.CRS:
            crs = {"init": "epsg:%s" % inD.crs.to_epsg()}
        else:
            crs = inD.crs
        cMeta = {
            "count": 1,
            "crs": crs,
            "dtype": inD["VALUE"].dtype,
            "driver": "GTiff",
            "transform": nTransform,
            "height": height,
            "width": width,
            "nodata": nodata,
        }
    shapes = ((row.geometry, row.VALUE) for idx, row in inD.iterrows())
    burned = features.rasterize(
        shapes=shapes,
        out_shape=(cMeta["height"], cMeta["width"]),
        transform=nTransform,
        dtype=cMeta["dtype"],
        merge_alg=mAlg,
        fill=nodata,
    )
    if outFile:
        try:
            with rasterio.open(outFile, "w", **cMeta) as out:
                out.write_band(1, burned)
        except:
            print("Error writing raster")
    return {"meta": cMeta, "vals": burned}


def polygonizeArray(data, curRaster):
    """Convert input array (data) to a geodataframe

    :param data: numpy array of raster data. ie - rasterio.open().read()
    :type data: np.array
    :param curRaster: template raster object
    :type curRaster: rasterio.DatasetReader
    :return: geodataframe with columns row, col, val, geometry
    :rtype: gpd.GeoDataFrame
    """
    # Calculate resolution of cells
    b = curRaster.bounds
    ll = curRaster.xy(*curRaster.index(*b[0:2]), "ll")
    xmin = ll[0]
    ymin = ll[1]
    xRes = curRaster.res[0]
    yRes = curRaster.res[1]
    crs = curRaster.crs
    # create a dataframe equal to the size of the array
    outArray = pd.DataFrame()
    outArray["id"] = list(range(0, (data.shape[0] * data.shape[1])))
    rowVals = []
    colVals = []
    actualvals = []
    for row in range(0, data.shape[0]):
        for col in range(0, data.shape[1]):
            rowVals.append(row)
            colVals.append(col)
            actualvals.append(data[row, col])
    outArray["row"] = rowVals
    outArray["col"] = colVals
    outArray["vals"] = actualvals

    # Create a polygon covering each cell
    def getPolygon(x):
        llX = xmin + (xRes * x["col"])
        llY = ymin + (yRes * x["row"])
        A = "%s %s" % (llX, llY)
        B = "%s %s" % (llX, llY + yRes)
        C = "%s %s" % (llX + xRes, llY + yRes)
        D = "%s %s" % (llX + xRes, llY)
        return wkt.loads("POLYGON((%s,%s,%s,%s,%s))" % (A, B, C, D, A))

    outArray["geometry"] = outArray.apply(getPolygon, axis=1)
    outGeo = gpd.GeoDataFrame(outArray, geometry="geometry")
    outGeo.crs = crs
    return outGeo


def zonalStats(
    inShp,
    inRaster,
    bandNum=1,
    mask_A=None,
    reProj=False,
    minVal="",
    maxVal="",
    verbose=False,
    rastType="N",
    unqVals=[],
    weighted=False,
    allTouched=False,
    calc_sd=False,
    return_df=False,
):
    """Run zonal statistics against an input shapefile. Returns array of SUM, MIN, MAX, and MEAN

    :param inShp: input geospatial data to summarize raster
    :type inShp: string path to file of gpd.GeoDataFrame
    :param inRaster: input raster to summarize
    :type inRaster: string path to file or rasterio.DatasetReader
    :param bandNum: band in raster to analyze, defaults to 1
    :type bandNum: int, optional
    :param mask_A: mask the raster data using an identical shaped boolean mask, defaults to None
    :type mask_A: np.array, optional
    :param reProj: whether to reproject data to match, if not, raise a ValueError if CRS mismatch between inShp and inRaster, defaults to False
    :type reProj: bool, optional
    :param minVal: if defined, will only calculate statistics on values above this number, defaults to ''
    :type minVal: number, optional
    :param maxVal: if defined, will only calculate statistics on values below this number, defaults to ''
    :type maxVal: number, optional
    :param verbose: provide additional text updates, defaults to False
    :type verbose: bool, optional
    :param rastType: Type of raster, defaults to 'N' as numerical or 'C' as categorical. If 'C' is used, you should provide unqVals
    :type rastType: str, optional
    :param unqVals: List of unique values to search for in raster, defaults to []
    :type unqVals: list of int, optional
    :param weighted: apply weighted zonal calculations. This will determine the % overlap for each
        raster cell in the defined AOI. Will apply weights in calculations of numerical statistics, defaults to False
    :type weighted: bool, optional
    :param allTouched: whether to include all cells touched in raster calculation, passed to rasterio rasterize function, defaults to False
    :type allTouched: bool, optional
    :param calc_sd: include the standard deviation in calculation, defaults to False
    :type calc_sd: bool, optional
    :param return_df: if true, return result as data frame; defaults to False
    :type return_df: boolean, optional
    :raises ValueError: If CRS mismatch between inShp and inRaster
    :return: array of zonal results - one entry for every feature in inShp. Each entry is SUM, MIN, MAX, MEAN, SD (optional)
    :rtype: array
    """
    if isinstance(inShp, str):
        inVector = gpd.read_file(inShp)
    else:
        inVector = inShp
    if isinstance(inRaster, str):
        curRaster = rasterio.open(inRaster, "r")
    else:
        curRaster = inRaster

    # If mask is not none, apply mask
    if mask_A is not None:
        curRaster.write_mask(mask_A)

    outputData = []
    if inVector.crs != curRaster.crs:
        if reProj:
            inVector = inVector.to_crs(curRaster.crs)
        else:
            raise ValueError("Input CRS do not match")
    fCount = 0
    tCount = len(inVector["geometry"])
    # generate bounding box geometry for raster bbox
    b = curRaster.bounds
    rBox = box(b[0], b[1], b[2], b[3])
    for idx, row in inVector.iterrows():
        geometry = row["geometry"]
        fCount = fCount + 1
        try:
            # This test is used in case the geometry extends beyond the edge of the raster
            #   I think it is computationally heavy, but I don't know of an easier way to do it
            if not rBox.contains(geometry):
                geometry = geometry.intersection(rBox)
            try:
                if fCount % 1000 == 0 and verbose:
                    tPrint("Processing %s of %s" % (fCount, tCount))
                # get pixel coordinates of the geometry's bounding box
                ul = curRaster.index(*geometry.bounds[0:2])
                lr = curRaster.index(*geometry.bounds[2:4])
                # read the subset of the data into a numpy array
                window = (
                    (float(lr[0]), float(ul[0] + 1)),
                    (float(ul[1]), float(lr[1] + 1)),
                )

                if mask_A is not None:
                    data = curRaster.read(bandNum, window=window, masked=True)
                else:
                    data = curRaster.read(bandNum, window=window, masked=False)

                if weighted:
                    allTouched = True
                    # Create a grid of the input raster (data)
                    rGrid = polygonizeArray(data, geometry.bounds, curRaster)
                    # Clip the grid by the input geometry
                    rGrid["gArea"] = rGrid.area
                    rGrid["newArea"] = rGrid.intersection(geometry).area
                    # Store the percent overlap
                    rGrid["w"] = rGrid["newArea"] / rGrid["gArea"]
                    newData = data
                    for idx, row in rGrid.iterrows():
                        newData[row["row"], row["col"]] = (
                            data[row["row"], row["col"]] * row["w"]
                        )
                    data = newData

                """
                # Mask out no-data in data array
                if 'nodata' in curRaster.profile.keys():
                    no_data_val = curRaster.profile['nodata']
                    #data[data == no_data_val] = np.nan
                    data[data == no_data_val] = 0
                """

                # create an affine transform for the subset data
                t = curRaster.transform
                shifted_affine = Affine(
                    t.a, t.b, t.c + ul[1] * t.a, t.d, t.e, t.f + lr[0] * t.e
                )

                # rasterize the geometry
                mask = rasterize(
                    [(geometry, 0)],
                    out_shape=data.shape,
                    transform=shifted_affine,
                    fill=1,
                    all_touched=allTouched,
                    dtype=np.uint8,
                )

                # create a masked numpy array
                masked_data = np.ma.array(data=data, mask=mask.astype(bool))
                if rastType == "N":
                    if minVal != "" or maxVal != "":
                        if minVal != "":
                            masked_data = np.ma.masked_where(
                                masked_data < minVal, masked_data
                            )
                        if maxVal != "":
                            masked_data = np.ma.masked_where(
                                masked_data > maxVal, masked_data
                            )
                        if masked_data.count() > 0:
                            results = [
                                np.nansum(masked_data),
                                np.nanmin(masked_data),
                                np.nanmax(masked_data),
                                np.nanmean(masked_data),
                            ]
                        else:
                            results = [-1, -1, -1, -1]
                    else:
                        results = [
                            np.nansum(masked_data),
                            np.nanmin(masked_data),
                            np.nanmax(masked_data),
                            np.nanmean(masked_data),
                        ]
                    if calc_sd:
                        try:
                            results.append(np.std(masked_data))
                        except:
                            results.append(-1)
                if rastType == "C":
                    if len(unqVals) > 0:
                        masked_unq = np.unique(masked_data, return_counts=True)
                        xx = dict(
                            [
                                x
                                for x in list(zip(masked_unq[0], masked_unq[1]))
                                if x[0] != "masked"
                            ]
                        )
                        results = [xx.get(i, 0) for i in unqVals]
                    else:
                        results = np.unique(masked_data, return_counts=True)
                outputData.append(results)
            except Exception as e:
                if verbose:
                    print(e)
                if rastType == "N":
                    outputData.append([-1, -1, -1, -1])
                else:
                    outputData.append([-1 for x in unqVals])
        except:
            print("Error processing %s" % fCount)
    if return_df:
        cols = ["SUM", "MIN", "MAX", "MEAN"]
        if calc_sd:
            cols.append("SD")
        outputData = pd.DataFrame(outputData, columns=cols)
    return outputData


def standardizeInputRasters(inR1, inR2, inR1_outFile="", resampling_type="nearest"):
    """Standardize inR1 to inR2: changes crs, extent, and resolution.

    :param inR1: rasterio object for raster to be modified
    :type inR1: ratserio.DatasetReader
    :param inR2: rasterio object to be standardized to
    :type inR12 ratserio.DatasetReader
    :param inR1_outfile: path to create output raster file of standardized inR1, default is '', which means nothing is written
    :type inR1: string
    :param resampling_type: how to perfrom spatial resampling; options are nearest (default), cubic, or sum
    :type resampling_type: string
    :return: array of numpy array, and rasterio metadata
    :rtype: array
    """
    if inR1.crs != inR2.crs:
        bounds = gpd.GeoDataFrame(
            pd.DataFrame([[1, box(*inR2.bounds)]], columns=["ID", "geometry"]),
            geometry="geometry",
            crs=inR2.crs,
        )
        bounds = bounds.to_crs(inR1.crs)
        b2 = bounds.total_bounds
        boxJSON = [
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [b2[0], b2[1]],
                        [b2[0], b2[3]],
                        [b2[2], b2[3]],
                        [b2[2], b2[1]],
                        [b2[0], b2[1]],
                    ]
                ],
            }
        ]
    else:
        b2 = inR2.bounds
        boxJSON = [
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [b2.left, b2.bottom],
                        [b2.left, b2.top],
                        [b2.right, b2.top],
                        [b2.right, b2.bottom],
                        [b2.left, b2.bottom],
                    ]
                ],
            }
        ]
    # Clip R1 to R2
    # Get JSON of bounding box
    out_img, out_transform = mask(inR1, boxJSON, crop=True)
    out_img[out_img < 0] = 0
    out_meta = inR1.meta.copy()
    # Re-scale resolution of R1 to R2
    newArr = np.empty(shape=(1, inR2.shape[0], inR2.shape[1]))

    if resampling_type == "cubic":
        resampling_type = Resampling.cubic
    elif resampling_type == "nearest":
        resampling_type = Resampling.nearest
    elif resampling_type == "sum":
        resampling_type = Resampling.sum
    reproject(
        out_img,
        newArr,
        src_transform=out_transform,
        dst_transform=inR2.transform,
        src_crs=inR1.crs,
        dst_crs=inR2.crs,
        resampling=resampling_type,
    )
    out_meta.update(
        {
            "driver": "GTiff",
            "height": newArr.shape[1],
            "width": newArr.shape[2],
            "transform": inR2.transform,
            "crs": inR2.crs,
        }
    )
    if inR1_outFile != "":
        with rasterio.open(inR1_outFile, "w", **out_meta) as dest:
            dest.write(newArr.astype(out_meta["dtype"]))
    return [newArr.astype(out_meta["dtype"]), out_meta]


def jaccardIndex(inR1, inR2):
    """Calculate the jaccard index on two binary input raster objects; Reference: https://en.wikipedia.org/wiki/Jaccard_index

    :param inR1: binary rasterio raster object to compare; needs to be same shape as inR2
    :type inR1: rasterio.DatasetReader
    :param inR2: binary rasterio raster object to compare; needs to be same shape as inR1
    :type inR2: rasterio.DatasetReader
    :raises ValueError: if inR1 and inR2 are different shapes
    :return: index comparing similarity of input raster datasets
    :rtype: float
    """
    if inR1.shape != inR2.shape:
        print(inR1.shape)
        print(inR2.shape)
        raise ValueError("Shape of input rasters do not match")
    # Add the two rasters together and get the unique tabulation
    inC = inR1.read() + inR2.read()
    xx = np.unique(inC, return_counts=True)
    outDict = {}
    for itemIdx in range(0, len(xx[0])):
        outDict[xx[0][itemIdx]] = xx[1][itemIdx]

    # The resulting could have some weird numbers, but values 1 and 2 should be the focus.
    #   1 - Only one area defines it as urban
    #   2 - Both areas define cell as urban
    # Jaccard is ratio of 2 / 1+2
    try:
        jIdx = outDict[2] / float(outDict[2] + outDict[1])
        return jIdx
    except:
        return -1



"""
The following module contains a number of functions to aggregate geospatial outputs into tables for InfraSAP analytics.
"""
import geopandas as gpd
import pandas as pd
import rasterio as rio
from rasterio import features
from rasterstats import zonal_stats
import numpy as np
from shapely.wkt import loads


def rasterize_gdf(inD, field, template, outFile=None, nodata=np.nan):
    """Convert geopandas geo data frame to raster of equal size/res to template raster

    INPUT
    inD [ geopandas data frame / path ]
    outFile [ string ] - path to save output raster
    field [ string ] - field to rasterize
    template [ string ] - path to template raster
    nodata [ int ] - value for no data

    RETURNS
    Raster file. If no outFile is specified, the function returns burned features as numpy array
    """
    raster_template = rio.open(template)
    # get info from template file
    xRes = raster_template.res[1]
    yRes = raster_template.res[0]
    trans = raster_template.transform
    x_pixels = raster_template.shape[1]
    y_pixels = raster_template.shape[0]
    raster_template.close()

    shapes = ((row.geometry, row[field]) for idx, row in inD.iterrows())
    burned = features.rasterize(
        shapes=shapes, fill=nodata, out_shape=raster_template.shape, transform=trans
    )
    burned = burned.astype(str(inD[field].dtype))

    if outFile:
        with rio.open(
            outFile,
            "w",
            driver="GTiff",
            height=y_pixels,
            width=x_pixels,
            count=1,
            dtype=str(inD[field].dtype),
            crs=raster_template.crs,
            transform=trans,
            nodata=nodata,
        ) as new_dataset:
            new_dataset.write_band(1, burned)
    else:
        return burned


def pop_weighted_average(target, data_raster, pop_raster, new_field, table=None):
    """Calculate population weighted average from a raster dataset to a target shapefile

    INPUT
    target [ GDF ] - Target geo data frame
    data_raster [ path ]
    pop_raster [ path ]
    new_field [ string ] -

    RETURNS
    Target GDF with new fields
    """
    # Load inputs
    if type(target) == str:
        target = gpd.read_file(target)

    pop = rio.open(pop_raster)
    pop_crs = pop.crs.to_string()
    pop_array = pop.read(1, masked=True)

    if pop_crs != target.crs.to_string():
        target = target.to_crs(pop_crs)

    # Calculate weights
    zs_sum_pop = pd.DataFrame(
        zonal_stats(
            target, pop_array, affine=pop.transform, stats="sum", nodata=pop.nodata
        )
    ).rename(columns={"sum": "pop_sum"})
    target_temp = target.join(zs_sum_pop)
    pop_sum_array = rasterize_gdf(inD=target_temp, field="pop_sum", template=pop_raster)
    weights = pop_array / pop_sum_array

    # Apply weights
    data = rio.open(data_raster).read(1, masked=True)
    data_weighted = weights * data

    zs_sum_data = pd.DataFrame(
        zonal_stats(
            target,
            data_weighted.filled(),
            affine=pop.transform,
            stats="sum",
            nodata=pop.nodata,
        )
    ).rename(columns={"sum": new_field})

    if table is not None:
        table = table.join(zs_sum_data)
        return table
    else:
        target = target.join(zs_sum_data)
        return target


def pop_weighted_average_national(data_raster, pop_raster, label, table=None):
    """Calculate population weighted average from a raster dataset to a target shapefile

    INPUT
    data_raster [ path ]
    pop_raster [ path ]
    new_field [ string ] -

    RETURNS
    Target GDF with new fields
    """
    pop = rio.open(pop_raster)
    pop_crs = pop.crs.to_string()
    pop_array = pop.read(1, masked=True)

    # Calculate weights
    pop_sum_array = pop_array.sum()
    weights = pop_array / pop_sum_array

    # Apply weights
    data = rio.open(data_raster).read(1, masked=True)
    data_weighted = weights * data

    sum_data = data_weighted.sum()
    results = pd.DataFrame(
        data=[sum_data], index=[label], columns=["pop weighted average"]
    )
    if table is not None:
        table = pd.concat([table, results])
        return table
    else:
        return results


def calculate_access_percentages(
    OD,
    target,
    dest_type,
    rural=False,
    urban_extents=None,
    pop_threshold=None,
    thresholds=[0, 30, 60, 120, 180, 240, 300, 360, 2000],
    capital=None,
):
    if type(target) == str:
        target = gpd.read_file(target).reset_index(drop=True)
    if type(OD) == str:
        OD = pd.read_csv(OD, header=[0, 1], index_col=0)
    if type(urban_extents) == str:
        urban_extents = gpd.read_file(urban_extents)

    origins_geom = OD["origin"][["geometry"]].copy()
    origins_geom.loc[:, "geometry"] = origins_geom["geometry"].apply(lambda x: loads(x))
    origins_geom = gpd.GeoDataFrame(origins_geom, crs="EPSG:4326", geometry="geometry")

    origins_sj = gpd.sjoin(
        origins_geom, target[["geometry"]], how="left", op="intersects"
    )
    OD.loc[:, ("origin", "target_idx")] = origins_sj["index_right"]

    if rural:
        origins_sj2 = gpd.sjoin(
            origins_geom, urban_extents, how="left", op="intersects"
        )
        OD.loc[:, ("origin", "rural")] = origins_sj2["index_right"].apply(
            lambda x: 1 if pd.isna(x) else 0
        )
        OD = OD.loc[OD[("origin", "rural")] == 1].copy()

    OD_dest = OD[dest_type]
    if pop_threshold:
        major_cities = [
            str(x) for x in urban_extents.loc[urban_extents.Pop >= pop_threshold].index
        ]
        OD_dest = OD_dest.loc[:, major_cities]
    if capital:
        OD_dest = OD_dest.loc[:, [capital]]

    min_df = OD["origin"].join(
        pd.DataFrame(OD_dest.min(axis=1), columns=["tt_min"])
    )  # .apply(lambda x: (x/60))
    min_df = min_df.loc[~pd.isna(min_df.target_idx)]
    min_df.loc[:, "target_idx"] = min_df.target_idx.astype(int)
    min_df.loc[:, "tt_min_cut"] = pd.cut(min_df.tt_min, bins=thresholds)

    summary = (
        min_df.groupby(["target_idx", "tt_min_cut"])[["pointid"]]
        .sum()
        .unstack()
        .fillna(0)
    )
    summary.columns = summary.columns.get_level_values(1)
    summary_pct = summary.apply(lambda x: x / (summary.sum(axis=1))).fillna(0)
    #     summary_pct.columns = summary_pct.columns.get_level_values(1)
    results = target.join(summary_pct).join(summary, rsuffix=" pop")

    return results


def agregate_to_country(summary, indicator, table=None):
    pop_cols = summary.columns[["pop" in x for x in summary.columns]]
    national = summary[pop_cols].sum(axis=0)
    national_pct = pd.DataFrame(
        national.apply(lambda x: x / (national.sum())), columns=[indicator]
    ).transpose()

    if table is not None:
        table = pd.concat([table, national_pct])
        return table
    else:
        return national_pct



import contextily as ctx
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import matplotlib.patches as mpatches

from matplotlib.patches import Patch


def static_map_vector(
    v_data,
    map_column,
    colormap="Reds",
    edgecolor="darker",
    reverse_colormap=False,
    thresh=None,
    legend_loc="upper right",
    figsize=(10, 10),
    out_file="",
    set_title=True,
):
    """Simple plot of vector data; most arguments expect

    :param v_data: input geopandas dataset to map
    :type v_data: gpd.GeoDataFrame
    :param map_column: Column label in v_data to map
    :type map_column: str
    :param colormap: Name of colour ramp to send to matplotlib.pyplot, defaults to "Reds"
    :type colormap: str, optional
    :param edgecolor: Optional parameter to change edge colour of polygons.
                      Optional values are match, darker, or a single provided colour, defaults to 'darker'
    :type edgecolor: str, optional
    :param reverse_colormap: Optionally reverse the colormap colorramp, defaults to False
    :type reverse_colormap: bool, optional
    :param thresh: List of thresholds to categorize values in v_data[map_column], defaults to equal interval 6 classes
    :type thresh: List of int, optional
    :param legend_loc: Where to place legend in plot, plugs into ax.legend, defaults to "upper right"
    :type legend_loc: str, optional
    :param figsize: Size of image, defaults to (10,10)
    :type figsize: tuple, optional
    :param out_file: path to create output image, defaults to '', which creates no output file
    :type out_file: str, optional
    :return: matplotlib object containing all maps
    :rtype: matplotlib.pyplot
    """
    geom_type = v_data["geometry"].geom_type.iloc[0]

    if v_data.crs.to_epsg() != 3857:
        v_data = v_data.to_crs(3857)
    # classify the data into categories if threshold is defined
    try:
        if thresh:
            v_data["tomap"] = pd.cut(
                v_data[map_column], thresh, labels=list(range(0, len(thresh) - 1))
            )
        else:
            v_data["tomap"] = pd.cut(v_data[map_column], 6, labels=[0, 1, 2, 3, 4, 5])
    except:
        print("Error mapping specified column, defaulting to index")
        map_column = "fake_index"
        v_data[map_column] = list(v_data.index)
        v_data["tomap"] = pd.cut(v_data[map_column], 6, labels=[0, 1, 2, 3, 4, 5])
    fig, ax = plt.subplots(figsize=figsize)
    cm = plt.cm.get_cmap(colormap)
    if reverse_colormap:
        cm = cm.reversed()
    all_labels = []
    for label, mdata in v_data.groupby("tomap"):
        if mdata.shape[0] > 0:
            color = cm(label / v_data["tomap"].max())
            # Determine edge color
            if edgecolor == "match":
                c_edge = color
            elif edgecolor == "darker":
                c_edge = (color[0] * 0.7, color[1] * 0.7, color[2] * 0.7, color[3])
            else:
                c_edge = edgecolor

            if geom_type == "Point":
                mdata.plot(
                    color=color, ax=ax, label=label, edgecolor=c_edge, markersize=300
                )
            elif geom_type == "Polygon":
                mdata.plot(color=color, ax=ax, label=label, edgecolor=c_edge)
            else:  # should handle lines; not yet tested
                mdata.plot(color=color, ax=ax, label=label, edgecolor=c_edge)
            try:
                cLabel = f"{round(mdata[map_column].min())} - {round(mdata[map_column].max())}"
            except:
                cLabel = "LABEL"
            cur_patch = mpatches.Patch(color=color, label=cLabel)
            all_labels.append(cur_patch)
    try:
        ctx.add_basemap(
            ax, source=ctx.providers.Stamen.TonerBackground
        )  # zorder=-10, 'EPSG:4326'
    except:
        print("Error adding basemap")
    ax.legend(handles=all_labels, loc=legend_loc)
    ax = ax.set_axis_off()

    if set_title:
        plt.title(map_column)

    if out_file != "":
        plt.savefig(out_file, dpi=300, bbox_inches="tight")

    return plt


def static_map_raster(
    r_data,
    colormap="magma",
    reverse_colormap=False,
    thresh=None,
    legend_loc="upper right",
    figsize=(10, 10),
    out_file="",
):
    """Simple plot of raster data

    :param r_data: Raster data to map, plots the first bad in the raster dataset
    :type r_data: rasterio.RasterDatasetReader
    :param colormap: Name of colour ramp to send to matplotlib.pyplot, defaults to "Reds"
    :type colormap: str, optional
    :param reverse_colormap: Optionally reverse the colormap colorramp, defaults to False
    :type reverse_colormap: bool, optional
    :param thresh: List of thresholds to categorize values in v_data[map_column], defaults to equal interval 6 classes
    :type thresh: List of int, optional
    :param legend_loc: Where to place legend in plot, plugs into ax.legend, defaults to "upper right"
    :type legend_loc: str, optional
    :param figsize: Size of image, defaults to (10,10)
    :type figsize: tuple, optional
    :param vector_mask: _description_, defaults to None
    :type vector_mask: _type_, optional
    :param out_file: path to create output image, defaults to '', which creates no output file
    :type out_file: str, optional
    :return: matplotlib object containing all maps
    :rtype: matplotlib.pyplot
    """

    map_data = r_data.read()[0, :, :]
    map_data = np.nan_to_num(map_data, neginf=0, posinf=2000)
    cm = plt.cm.get_cmap(colormap)
    if reverse_colormap:
        cm = cm.reversed()

    if thresh:
        map_data = np.digitize(map_data, thresh)
    fig, ax = plt.subplots(figsize=figsize)
    chm_plot = ax.imshow(map_data, cmap=cm)

    legend_labels = [[cm(0), "Low"], [cm(0.5), "Medium"], [cm(1), "High"]]
    if thresh:
        legend_labels = [[cm(x / max(thresh)), str(x)] for x in thresh]

    patches = [Patch(color=x[0], label=x[1]) for x in legend_labels]
    ax.legend(handles=patches, loc=legend_loc, facecolor="white")
    ax.set_axis_off()
    if out_file != "":
        plt.savefig(out_file, dpi=300, bbox_inches="tight")
    return plt



import os

import geopandas as gpd
from geographiclib.geodesic import Geodesic
from shapely.geometry import Point, LineString


def get_centroid(iso, boundaries):
    try:
        selected_country = boundaries.loc[boundaries["ISO3"] == iso]
        if selected_country.shape[0] == 1:
            return selected_country.iloc[0]["geometry"]
        elif selected_country.shape[0] > 1:
            selected_country = selected_country.sort_values(
                "Shape_Area", ascending=False
            )
            return selected_country.iloc[0]["geometry"]
        else:
            return None
    except:
        return None


def generate_great_circle(from_pt, to_pt, interim_steps=15):
    """ """
    geod = Geodesic.WGS84
    g = geod.Inverse(from_pt.x, from_pt.y, to_pt.x, to_pt.y)
    l = geod.Line(g["lat1"], g["lon1"], g["azi1"])
    num = interim_steps  # 15 intermediate steps
    list_of_points = [from_pt]
    for i in range(num + 1):
        pos = l.Position(i * g["s12"] / num)
        list_of_points.append(Point(pos["lat2"], pos["lon2"]))
    list_of_points.append(to_pt)
    return LineString(list_of_points)


def generate_line_string(row, type="normal"):
    try:
        if type == "normal":
            if row["Trade Flow"] == "Export":
                return LineString([row["Reporter_Pt"], row["Partner_Pt"]])
            else:
                return LineString([row["Partner_Pt"], row["Reporter_Pt"]])
        elif type == "great":
            if row["Trade Flow"] == "Export":
                return generate_great_circle(row["Reporter_Pt"], row["Partner_Pt"])
            else:
                return generate_great_circle(row["Partner_Pt"], row["Reporter_Pt"])

    except:
        # print("Error processing %s and %s" % (row['Reporter ISO'], row['Partner ISO']))
        return None


class comtrade_flow(object):
    """Used to process comtrade flows of energy (etc.) trade flows between countries

    :param: comtrade_csv - string path to the csv containing the raw comtrade results
    :param: type - string describing the type of comtrade

    """

    def __init__(
        self,
        comtrade_csv,
        type,
        good_columns=[
            "Qty Unit Code",
            "Year",
            "Trade Flow",
            "Reporter ISO",
            "Partner ISO",
            "Commodity",
            "Qty",
            "Trade Value (US$)",
            "Reporter_Pt",
            "Partner_Pt",
            "TOE",
        ],
    ):
        """
        initiates the comtrade_flow object, see help(comtrade_flow) for more information
        """
        self.type = type
        self.csv_file = comtrade_csv
        self.type = type
        self.good_columns = good_columns
        self.raw_data = gpd.read_file(self.csv_file)
        self.raw_data.crs = {"init": "epsg:4326"}

    def initialize(self, good_quantity_code, inB, line_type="normal"):
        """
        Read in comtrade data, filter data without units, etc.

        :param: good_quantity_code - array of numbers describing the values in the column 'Qty Unit Code' that are acceptable
        :param: inB - geopandas data frame of national centroids. Must contain columns "ISO3" and "geometry"
        :param: line_type - connection type between flows. Can be set to "great" to generate great circles
        """

        good_quantity_code = [str(x) for x in good_quantity_code]
        inD = self.raw_data
        for n_field in ["Trade Value (US$)", "Qty", "TOE"]:
            try:
                inD[n_field] = inD[n_field].astype(float)
            except:
                inD[n_field] = inD[n_field].replace("", "0").astype(float)
            try:
                inD[n_field] = inD[n_field].astype(float)
            except:
                inD[n_field] = inD[n_field].replace("", "0").astype(float)
        # filter out bad data
        if len(good_quantity_code) > 0:
            inD = inD.loc[
                inD["Qty Unit Code"].isin(good_quantity_code), self.good_columns
            ]

        # Collapse commodities into single rows
        inD = (
            inD.groupby(["Reporter ISO", "Partner ISO", "Year", "Trade Flow"])[
                "Qty", "Trade Value (US$)", "TOE"
            ]
            .sum()
            .reset_index()
        )

        inD["Reporter_Pt"] = inD["Reporter ISO"].apply(lambda x: get_centroid(x, inB))
        inD["Partner_Pt"] = inD["Partner ISO"].apply(lambda x: get_centroid(x, inB))

        # generate country flows geometry
        country_flows = inD.loc[inD["Partner ISO"] != "WLD"]
        country_flows["geometry"] = country_flows.apply(
            lambda x: generate_line_string(x, line_type), axis=1
        )
        country_summary = inD.loc[inD["Partner ISO"] == "WLD"]

        # store data
        self.complete_data = inD
        self.country_flows = country_flows
        self.country_summary = country_summary

    def save_simple_layers(self, out_folder, out_type="SHP"):
        """
        Extract the most important layers from the country_flows and country_summary
        and save to disk in order to put on GIM
        """
        if not os.path.exists(out_folder):
            os.makedirs(out_folder)

        self.clean_fields()
        self.country_summary["Value per unit"] = (
            self.country_summary["Trade Value (US$)"] / self.country_summary["Qty"]
        )
        self.country_flows["Value per unit"] = (
            self.country_flows["Trade Value (US$)"] / self.country_flows["Qty"]
        )

        # Summarize country summaries
        agg = self.country_summary.groupby(["Reporter ISO", "Trade Flow"])
        most_recent = agg.last().reset_index()
        most_recent_import = most_recent.loc[most_recent["Trade Flow"] == "Import"]
        most_recent_export = most_recent.loc[most_recent["Trade Flow"] == "Export"]
        # Summarize country summaries
        agg_flow = self.country_flows.groupby(
            ["Reporter ISO", "Partner ISO", "Trade Flow"]
        )
        most_recent_flow = agg_flow.last().reset_index()
        most_recent_import_flow = most_recent_flow.loc[
            most_recent_flow["Trade Flow"] == "Import"
        ]
        most_recent_export_flow = most_recent_flow.loc[
            most_recent_flow["Trade Flow"] == "Export"
        ]

        if out_type == "CSV":
            geom_driver = "CSV"
            out_type = "csv"
        elif out_type == "SHP":
            geom_driver = "ESRI Shapefile"
            out_type = "shp"
        elif out_type == "GEOJSON":
            geom_driver = "GeoJSON"
            out_type = "geojson"

        for fileDef in [
            [
                most_recent_import,
                os.path.join(out_folder, f"country_summary_imports.{out_type}"),
            ],
            [
                most_recent_export,
                os.path.join(out_folder, f"country_summary_exports.{out_type}"),
            ],
            [
                most_recent_import_flow,
                os.path.join(out_folder, f"country_flows_imports.{out_type}"),
            ],
            [
                most_recent_export_flow,
                os.path.join(out_folder, f"country_flows_exports.{out_type}"),
            ],
        ]:
            xx = fileDef[0]
            xx = gpd.GeoDataFrame(xx, geometry="geometry", crs={"init": "epsg:4326"})
            xx.to_file(fileDef[1], driver=geom_driver)

    def clean_fields(self):
        """ """
        # Convert geometry columns to text
        self.country_flows["Reporter_Pt"] = self.country_flows["Reporter_Pt"].apply(str)
        self.country_flows["Partner_Pt"] = self.country_flows["Partner_Pt"].apply(str)

        # Convert country summary columns to geometry
        if "Reporter_Pt" in self.country_summary.columns:
            country_summary_geom = self.country_summary["Reporter_Pt"]
            self.country_summary.drop(["Reporter_Pt"], axis=1, inplace=True)
            self.country_summary["geometry"] = country_summary_geom

    def save(self, out_folder, out_type="CSV"):
        """
        Save all the outputs to a folder

        :param: out_folder - string path to output folder; will be created if it does not exist
        :param: out_type - string to determine the format to save the results. Default is CSV, optionally "SHP"
        import fiona
        fiona.supported_drivers
        """
        if not os.path.exists(out_folder):
            os.makedirs(out_folder)

        self.clean_fields()

        if out_type == "CSV":
            geom_driver = "CSV"
            out_type = "csv"
            country_flows_file = os.path.join(out_folder, f"country_flows.{out_type}")
            country_summary_file = os.path.join(
                out_folder, f"country_summary.{out_type}"
            )
            self.country_flows.to_csv(country_flows_file)
            self.country_summary.to_csv(country_summary_file)
        if out_type == "SHP":
            geom_driver = "ESRI Shapefile"
            out_type = "shp"
            country_flows_file = os.path.join(out_folder, f"country_flows.{out_type}")
            country_summary_file = os.path.join(
                out_folder, f"country_summary.{out_type}"
            )
            self.country_flows.to_file(country_flows_file)
            self.country_summary.to_file(country_summary_file)



import json
import urllib
import boto3

import geopandas as gpd

from botocore.config import Config
from botocore import UNSIGNED


def aws_search_ntl(
    bucket="globalnightlight",
    prefix="composites",
    region="us-east-1",
    unsigned=True,
    verbose=False,
):
    """get list of nighttime lights files from open AWS bucket - https://registry.opendata.aws/wb-light-every-night/

    :param bucket: bucket to search for imagery, defaults to 'globalnightlight'
    :type bucket: str, optional
    :param prefix: prefix storing images. Not required for LEN, defaults to 'composites'
    :type prefix: str, optional
    :param region: AWS region for bucket, defaults to 'us-east-1'
    :type region: str, optional
    :param unsigned: if True, search buckets without stored boto credentials, defaults to True
    :type unsigned: bool, optional
    :param verbose: print additional support messages, defaults to False
    :type verbose: bool, optional
    """
    if unsigned:
        s3client = boto3.client("s3", config=Config(signature_version=UNSIGNED))
    else:
        s3client = boto3.client("s3")

    # Loop through the S3 bucket and get all the keys for files that are .tif
    more_results = True
    loops = 0
    good_res = []
    while more_results:
        if verbose:
            print(f"Completed loop: {loops}")
        if loops > 0:
            objects = s3client.list_objects_v2(
                Bucket=bucket, Prefix=prefix, ContinuationToken=token
            )
        else:
            objects = s3client.list_objects_v2(Bucket=bucket, Prefix=prefix)
        more_results = objects["IsTruncated"]
        if more_results:
            token = objects["NextContinuationToken"]
        loops += 1
        for res in objects["Contents"]:
            if res["Key"].endswith("avg_rade9.tif") and ("slcorr" in res["Key"]):
                good_res.append(
                    f"https://globalnightlight.s3.amazonaws.com/{res['Key']}"
                )

    return good_res


def get_geoboundaries(
    iso3,
    level,
    geo_api="https://www.geoboundaries.org/api/current/gbOpen/{iso3}/{adm}/",
):
    """Download boundaries dataset from geobounadries

    :param iso3: ISO3 code of country to download
    :type iso3: str
    :param level: Admin code to download in format of "ADM1" or "ADM2"
    :type level: str
    :return: spatial data representing the administrative boundaries
    :rtype: gpd.GeoDataFrame
    """
    cur_url = geo_api.format(iso3=iso3, adm=level)
    try:
        with urllib.request.urlopen(cur_url) as url:
            data = json.load(url)
            geo_data = gpd.read_file(data["gjDownloadURL"])
            return geo_data
    except:
        all_url = geo_api.format(iso3=iso3, adm="ALL")
        raise (
            ValueError(
                f"Cannot find admin dataset {cur_url}. Check out {all_url} for details on what is available"
            )
        )



import os
import rasterio
import affine

import geopandas as gpd

from rasterio import features
from shapely.geometry import box


class gsm_rasterizer(object):
    def __init__(self, gsm_files, out_folder):
        """Used to open, process, and extract GSM coverage files on a country by country basis

        :param gsm_files: list of string paths defining the GSM coverage data
        """
        raw_gsm = {}
        for f in gsm_files:
            gsm_name = os.path.basename(f).replace(".shp", "")
            raw_gsm[gsm_name] = f
        self.gsm_files = raw_gsm
        self.out_folder = out_folder
        if not os.path.exists(self.out_folder):
            os.makedirs(out_folder)

    def initial_read_in(self, gsm_files=[]):
        if len(gsm_files) == 0:
            gsm_files = self.gsm_files
        gsm_data = {}

        for key, value in gsm_files.items():
            gsm_data[key] = gpd.read_file(value)

        self.gsm_data = gsm_data

    def get_exact_shape(self, x, boundary):
        if boundary.contains(x):
            return x
        else:
            try:
                return boundary.intersection(x)
            except:
                x = x.buffer(0)
                return boundary.intersection(x)

    def extract_country_vectors(self, iso3, global_data, out_folder=""):
        country_data = global_data[global_data["ISO3"] == iso3]
        sample_key = list(self.gsm_data.keys())[0]
        if country_data.crs != self.gsm_data[sample_key].crs:
            country_data = country_data.to_crs(self.gsm_data[sample_key].crs)
        if out_folder == "":
            out_folder = self.out_folder
        self.country_data = country_data

        c_shp = box(*country_data.total_bounds)
        self.c_shp = c_shp
        out_gsm_files = []

        for key, gsm in self.gsm_data.items():
            out_file = os.path.join(out_folder, "%s.shp" % key)
            out_gsm_files.append(out_file)
            if not os.path.exists(out_file):
                select = gsm[gsm.intersects(c_shp)]
                if select.shape[0] > 0:
                    # clip geometries for features to actual boundary
                    select["geometry"] = select["geometry"].apply(
                        lambda x: self.get_exact_shape(x, c_shp)
                    )
                    select.to_file(out_file)
                else:
                    pass
        self.out_gsm_files = out_gsm_files

    def rasterize_gsm_vectors(self):
        for vec_file in self.out_gsm_files:
            if os.path.exists(vec_file):
                gsmD = gpd.read_file(vec_file)
                self.rasterize_gsm(gsmD, vec_file.replace(".shp", ".tif"))

    def rasterize_gsm(self, gsm, out_tif, cell_width=0.01):
        bounds = gsm.total_bounds
        width_cells = int(round((bounds[2] - bounds[0]) / cell_width))
        height_cells = int(round(((bounds[3] - bounds[1]) / cell_width)))
        cAffine = affine.Affine(cell_width, 0, bounds[0], 0, cell_width * -1, bounds[3])
        cMeta = {
            "count": 1,
            "crs": gsm.crs,
            "dtype": "uint8",
            "affine": cAffine,
            "driver": "GTiff",
            "transform": cAffine,
            "height": height_cells,
            "width": width_cells,
        }
        shapes = ((row.geometry, 1) for idx, row in gsm.iterrows())
        with rasterio.open(out_tif, "w", **cMeta) as out:
            burned = features.rasterize(
                shapes=shapes,
                fill=0.0,
                out_shape=(cMeta["height"], cMeta["width"]),
                transform=out.transform,
            )
            burned = burned.astype(cMeta["dtype"])
            out.write_band(1, burned)



import os
import logging

import rasterio  # v1.0.21 (latest v1.2.10)

import pandas as pd  # v1.0.3 (latest v1.4.1)
import geopandas as gpd  # v0.6.3 (latest v0.10.2)


from pyproj.crs.crs import CRS

try:
    pass
except:
    print("METADATA Library: Could not import arcgis libraries")

vector_file_types = [".shp", ".kml", ".geojson"]
raster_file_types = [".TIF", ".tif", ".tiff", ".geotiff", ".geotif"]


### TODO:
# 1. Add searching through ESRI geodatabases
# 2. Add support to include tables (xlsx, csv)


class metadata_gost:
    """Create standardized metadata for folders of geospatial data"""

    def __init__(self, input_dir, output_dir, type="Folder"):
        """Create standardized metadata for folders of geospatial data

        :param input_dir: Folder to search for geospatial layers
        :type input_dir: str
        :param output_dir: Where to create output metadata files
        :type output_dir: str
        :param type: What we are searching, defaults to "Folder"
        :type type: str, optional
        """
        self.input_dir = input_dir
        if not os.path.exists(input_dir):
            raise (ValueError("Input directory does not exist"))
        self.output_dir = output_dir
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

    def get_layers(self):
        """Iterate through self.input_dir and get list of vector and raster data"""
        vector_files = []
        raster_files = []

        for root, dirs, files in os.walk(self.input_dir):
            for f in files:
                file_type = os.path.splitext(f)[-1]
                if file_type in vector_file_types:
                    vector_files.append(os.path.join(root, f))
                if file_type in raster_file_types:
                    raster_files.append(os.path.join(root, f))
        self.vector_files = vector_files
        self.raster_files = raster_files
        return {"vector": vector_files, "raster": raster_files}

    def write_metadata(
        self,
        out_file,
        layer_metadata=None,
        field_metadata=None,
        dataset_id="",
        dataset_title="",
        country="",
        abstract="",
        purpose="",
        creation_date="",
        release_date="",
        owner="",
        email="",
    ):
        # create first sheet of baseline info
        base_info = {
            "dataset ID": dataset_id,
            "dataset title": dataset_title,
            "country": country,
            "abstract": abstract,
            "purpose": purpose,
            "creation date": creation_date,
            "release date": release_date,
            "owner name": owner,
            "owner email": email,
        }
        base_pd = pd.DataFrame([base_info]).transpose()
        if layer_metadata is None:
            layer_metadata = self.metaPD
        if field_metadata is None:
            field_metadata = self.fieldsPD

        with pd.ExcelWriter(
            out_file, engine="openpyxl", mode="w", if_sheet_exists="replace"
        ) as writer:
            base_pd.to_excel(writer, "dataset info", encoding="utf8")
            layer_metadata.to_excel(writer, "layer_summaries", encoding="utf8")
            field_metadata.to_excel(writer, "field_summaries", encoding="utf8")

    def generate_metadata(self, vector_files=None, raster_files=None):
        """Generate metadata for vector and raster files

        Args:
            vector_files [optional, list] - list of paths to vector_files, if none, defaults to list generated through get_list_of_layers
            raster_files [optional, list] - list of paths to raster_files, if none, defaults to list generated through get_list_of_layers

        Returns:
            [geopandas DataFrame]

        Details:
            Standard Fields
                layer_name - name of layer (filename without extension)
                data_type  - vector, raster, or table
                crs_name
                crs_code
                num_dimensions - Number of fields in dataset, or bands in raster dataset
                min_lon
                max_lon
                min_lat
                max_lat
            Vector Fields:
                vector_shape_type
                vector_object_count
                table_num_rows
            Raster Fields:
                raster_width
                raster_height

            Manual Input
                layer_label
                description
                source_name
                source_url
                data_process_summary
        """
        if vector_files is None:
            vector_files = self.vector_files
        if raster_files is None:
            raster_files = self.raster_files

        metadata = []
        field_defs = []
        for vector_file in vector_files:
            try:
                cur_meta = vector_file_metadata(vector_file, in_folder=self.input_dir)
                metadata.append(cur_meta.get_metadata())
                field_defs.append(cur_meta.get_field_summaries())
            except:
                logging.error(f"Cannot log {vector_file}")

        for raster_file in raster_files:
            try:
                cur_meta = raster_file_metadata(raster_file, in_folder=self.input_dir)
                metadata.append(cur_meta.get_metadata())
            except:
                logging.error(f"Cannot log {raster_file}")

        metaPD = pd.DataFrame(metadata)
        metaPD["layer_label"] = ""
        metaPD["description"] = ""
        metaPD["source_name"] = ""
        metaPD["source_url"] = ""
        metaPD["data_process_summary"] = ""
        self.metaPD = metaPD
        try:
            del final
        except:
            pass
        for cur_fields in field_defs:
            cur_pd = pd.DataFrame(cur_fields)
            try:
                final = final.append(cur_pd)
            except:
                final = cur_pd
        try:
            fieldsPD = final.reset_index()
            self.fieldsPD = fieldsPD
        except:
            fieldsPD = None
            self.fieldsPD = None
        self.metaPD.sort_values(["folder", "layer_name"], inplace=True)
        self.fieldsPD.sort_values(["layer_name", "field_name"], inplace=True)

        return {"metadata": metaPD, "fields": fieldsPD}


class raster_file_metadata:
    def __init__(self, path, in_folder=""):
        self.path = path
        self.layer_name = os.path.splitext((os.path.basename(path)))[0]
        if in_folder == "":
            self.dir_name = os.path.dirname(path)
        else:
            self.dir_name = os.path.dirname(path).replace(in_folder, "")
        self.data_type = "Raster"

        curR = rasterio.open(path)

        cur_crs = CRS.from_wkt(curR.crs.wkt)
        self.crs_name = cur_crs.name
        self.crs_code = curR.crs.to_epsg()

        rShape = curR.shape
        if len(rShape) == 2:
            self.num_dimensions = 1
            self.raster_width = curR.shape[0]
            self.raster_height = curR.shape[1]
        else:
            self.num_dimensions = curR.shape[0]
            self.raster_width = curR.shape[1]
            self.raster_height = curR.shape[2]

        self.raster_res = curR.res[0]
        b = curR.bounds
        self.min_lon = b[0]
        self.min_lat = b[1]
        self.max_lon = b[2]
        self.max_lat = b[3]

    def get_metadata(self):
        return {
            "layer_name": self.layer_name,
            "folder": self.dir_name,
            "data_type": self.data_type,
            "crs_name": self.crs_name,
            "crs_code": self.crs_code,
            "num_dimensions": self.num_dimensions,
            "min_lon": self.min_lon,
            "max_lon": self.max_lon,
            "min_lat": self.min_lat,
            "max_lat": self.max_lat,
            "raster_width": self.raster_width,
            "raster_height": self.raster_height,
            "raster_res": self.raster_res,
        }


class vector_file_metadata:
    def __init__(self, path, in_folder=""):
        self.path = path
        self.in_folder = ""
        self.layer_name = os.path.splitext((os.path.basename(path)))[0]
        if in_folder == "":
            self.dir_name = os.path.dirname(path)
        else:
            self.dir_name = os.path.dirname(path).replace(in_folder, "")
        self.data_type = "Vector"

        curD = gpd.read_file(path)
        self.curD = curD
        self.crs_name = curD.crs.name
        self.crs_code = curD.crs.to_epsg()
        self.num_dimensions = curD.shape[1]
        self.vector_object_count = curD.shape[0]

        b = curD.total_bounds
        self.min_lon = b[0]
        self.min_lat = b[1]
        self.max_lon = b[2]
        self.max_lat = b[3]
        self.vector_shape_type = curD.geom_type.value_counts().index[0]

    def get_field_summaries(self):
        """Generate field metadata
        layer_name
        field_name
        field_label
        definition
        domain
        type
        num_unique_values
        first_10_unique_values
        """
        cur_field_defs = []
        for idx, col in self.curD.iteritems():
            cur_defs = {
                "layer_name": self.layer_name,
                "field_name": idx,
                "type": col.dtype,
                "field_label": "",
                "definition": "",
                "domain": "",
                "num_unique_values": len(col.unique()),
                "first_10_unique_values": col.values[:10],
            }
            cur_field_defs.append(cur_defs)
        return cur_field_defs

    def get_metadata(self):
        return {
            "layer_name": self.layer_name,
            "data_type": self.data_type,
            "crs_name": self.crs_name,
            "crs_code": self.crs_code,
            "num_dimensions": self.num_dimensions,
            "min_lon": self.min_lon,
            "max_lon": self.max_lon,
            "min_lat": self.min_lat,
            "max_lat": self.max_lat,
            "vector_shape_type": self.vector_shape_type,
            "vector_object_count": self.vector_object_count,
        }



import time
import math
import logging

import geopandas as gpd
import pandas as pd
import numpy as np

from math import ceil
from shapely.geometry import Point, Polygon, box

from pyproj import CRS
from pyproj.aoi import AreaOfInterest
from pyproj.database import query_utm_crs_info


def loggingInfo(lvl=logging.INFO):
    """Set logging settings to info (default) and print useful information

    :param lvl: logging level for setting, defaults to logging.INFO
    :type lvl: logging.INFO
    """
    logging.basicConfig(format="%(asctime)s:%(levelname)s: %(message)s", level=lvl)


def tPrint(s):
    """prints the time along with the message"""
    print("%s\t%s" % (time.strftime("%H:%M:%S"), s))


def round_to_1(x):
    return round(x, -int(math.floor(math.log10(x))))


def drange(start, stop, step):
    """Create an interable range made with decimal point steps"""
    r = start
    while r < stop:
        yield r
        r += step


def getHistIndex(hIdx, val):
    """Get the index of a specific [val] within a list of histogram values

    :param hIdx: list of values (from histogram calculation)
    :type hIdx: list of numbers
    :param val: value to search for
    :type val: number
    :return: index in hIdx where val falls
    :rtype: int
    """
    lastH = 0
    for h in range(0, len(hIdx)):
        curH = hIdx[h]
        if curH > val:
            return lastH
        lastH = h
    return len(hIdx) - 1


def listSum(inD):
    """get sum of values in list

    :param inD: list of numbers
    :type inD: list
    """
    total = 0
    for x in inD:
        total += x
    return total


def getHistPer(inD):
    """Convert a list of values into a percent of total

    :param inD: list of values
    :type inD: list of numbers
    :return: list of values of same length of inD.
    :rtype: list of float
    """
    tSum = listSum(inD)
    for hIdx in range(0, len(inD)):
        inD[hIdx] = inD[hIdx] / tSum
    return inD


def tabulateUnq(unqResults, verbose=False, columnPrefix="c"):
    allVals = []
    for r in unqResults:
        allVals.append(r[0].tolist())
    flattened = [val for sublist in allVals for val in sublist]
    unq = np.unique(np.array(flattened)).tolist()
    # unqCols = ["c_%s" % xxx for xxx in unq]
    allRes = []
    for r in unqResults:
        try:
            curRes = [0] * len(unq)
            for idx in range(0, len(r[0].tolist())):
                curRes[unq.index(r[0].tolist()[idx])] = r[1].tolist()[idx]
        except:
            print(r)
        allRes.append(curRes)
    return pd.DataFrame(allRes, columns=["%s_%s" % (columnPrefix, xxx) for xxx in unq])


def createFishnet(
    xmin,
    xmax,
    ymin,
    ymax,
    gridHeight,
    gridWidth,
    type="POLY",
    crsNum=4326,
    outputGridfn="",
):
    """Create a fishnet shapefile inside the defined coordinates

    :param xmin: minimum longitude
    :type xmin: float
    :param xmax: maximum longitude
    :type xmax: float
    :param ymin: minimum latitude
    :type ymin: float
    :param ymax: maximum latitude
    :type ymax: float
    :param gridHeight: resolution of the grid cells in crsNum units
    :type gridHeight: float
    :param gridWidth: resolution of the grid cells in crsNum units
    :type gridWidth: float
    :param type: geometry type of output fishnet, defaults to 'POLY'
    :type type: str, optional
    :param crsNum: units of output crs, defaults to 4326
    :type crsNum: int, optional
    :param outputGridfn: path for output shapefile, defaults to '', which creates no shapefile
    :type outputGridfn: str, optional
    :return: geodataframe of fishnet
    :rtype: gpd.GeoDataFrame
    """

    def get_box(row, col, l, r, b, t, gridWidth, gridHeight):
        ll = Point(l + (row * gridWidth), b + (col + gridHeight))
        ul = Point(l + (row * gridWidth), t + (col + gridHeight))
        ur = Point(r + (row * gridWidth), t + (col + gridHeight))
        lr = Point(r + (row * gridWidth), b + (col + gridHeight))
        box = Polygon([ll, ul, ur, lr, ll])
        return box

    def get_point(row, col, l, r, b, t, gridWidth, gridHeight):
        pt = Point((l + r) / 2 + (col * gridWidth), (t + b) / 2 - (row * gridHeight))
        return pt

    # convert sys.argv to float
    xmin = float(xmin)
    xmax = float(xmax)

    ymin = float(ymin)
    ymax = float(ymax)
    gridWidth = float(gridWidth)
    gridHeight = float(gridHeight)

    # get rows
    rows = ceil((ymax - ymin) / gridHeight)
    # get columns
    cols = ceil((xmax - xmin) / gridWidth)

    # start grid cell envelope
    ringXleftOrigin = xmin
    ringXrightOrigin = xmin + gridWidth
    ringYtopOrigin = ymax
    ringYbottomOrigin = ymax - gridHeight

    all_res = []
    for rowIdx in range(0, rows):
        for colIdx in range(0, cols):
            if type == "POLY":
                box = get_box(
                    rowIdx,
                    colIdx,
                    ringXleftOrigin,
                    ringXrightOrigin,
                    ringYbottomOrigin,
                    ringYtopOrigin,
                    gridWidth,
                    gridHeight,
                )
            elif type == "POINT":
                box = get_point(
                    rowIdx,
                    colIdx,
                    ringXleftOrigin,
                    ringXrightOrigin,
                    ringYbottomOrigin,
                    ringYtopOrigin,
                    gridWidth,
                    gridHeight,
                )
            all_res.append([rowIdx, colIdx, box])
    res = gpd.GeoDataFrame(
        pd.DataFrame(all_res, columns=["rowIdx", "colIdx", "geometry"]),
        geometry="geometry",
        crs=f"epsg:{crsNum}",
    )
    if outputGridfn != "":
        res.to_file(outputGridfn)
    return res


def explodeGDF(indf):
    """Convert geodataframe with multi-part polygons to one with single part polygons

    :param indf: input geodataframe to explode
    :type indf: gpd.GeoDataFrame
    :return: exploded geodaatframe
    :rtype: gpd.GeoDataFrame
    """
    outdf = gpd.GeoDataFrame(columns=indf.columns)
    outdf.crs = indf.crs
    for idx, row in indf.iterrows():
        row.geometry.type
        if row.geometry.type in ["Polygon", "Point"]:
            outdf = outdf.append(row, ignore_index=True)
        if row.geometry.type in ["MultiPoint", "MultiPolygon"]:
            multdf = gpd.GeoDataFrame(columns=indf.columns)
            recs = len(row.geometry)
            multdf = multdf.append([row] * recs, ignore_index=True)
            for geom in range(recs):
                multdf.loc[geom, "geometry"] = row.geometry[geom]
            outdf = outdf.append(multdf, ignore_index=True)
    return outdf


def project_UTM(inD):
    """Project an input data frame to UTM coordinates.

    :param inD: input geodataframe to explode
    :type inD: gpd.GeoDataFrame
    :return: exploded geodaatframe
    :rtype: gpd.GeoDataFrame
    """
    import utm

    # get UTM zones for upper right and lower left of input dataframe

    if inD.crs != {"init": "epsg:4326"}:
        raise (ValueError("Cannot process input dataframe that is not WGS84"))

    inBounds = inD.total_bounds
    ll_utm = utm.from_latlon(inBounds[1], inBounds[0])
    ur_utm = utm.from_latlon(inBounds[3], inBounds[2])

    if (ll_utm[2] != ur_utm[2]) or (ll_utm[3] != ur_utm[3]):
        raise (
            ValueError(
                "The input shape spans multiple UTM zones: %s_%s to %s_%s"
                % (ll_utm[2], ll_utm[3], ur_utm[2], ur_utm[3])
            )
        )

    letter = "6"
    if ll_utm[3] == "U":
        letter = "7"
    outUTM = "32%s%s" % (letter, ll_utm[2])
    return inD.to_crs({"init": "epsg:%s" % outUTM})


def get_utm(gdf, bbox=True):
    if bbox:
        center = box(*gdf.total_bounds).centroid
    else:
        center = gdf.unary_union.centroid
    utm_crs_list = query_utm_crs_info(
        datum_name="WGS 84",
        area_of_interest=AreaOfInterest(
            west_lon_degree=center.x,
            south_lat_degree=center.y,
            east_lon_degree=center.x,
            north_lat_degree=center.y,
        ),
    )
    utm_crs = CRS.from_epsg(utm_crs_list[0].code)
    return utm_crs
